# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np
import scipy.special
import scipy.integrate
import matplotlib as mpl
import matplotlib.pyplot as plt

from mpl_toolkits    import mplot3d
from .nurbs_basis_functions    import compute_basis_polynomials, compute_basis_polynomials_derivatives


# -------------------------------------------------------------------------------------------------------------------- #
# Define the NURBS curve class
# -------------------------------------------------------------------------------------------------------------------- #
class NurbsCurve:

    """ Create a NURBS (Non-Uniform Rational Basis Spline) curve object

        Parameters
        ----------
        control_points : ndarray with shape (ndim, n+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points (0,1,...,n)

        weights : ndarray with shape (n+1,)
            Array containing the weight of the control points

        degree : int
            Degree of the basis polynomials

        knots : ndarray with shape (r+1=n+p+2,)
            The knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped NURBS


        Notes
        -----
        This class includes methods to compute:

            - Curve coordinates for any number of dimensions
            - Analytic curve derivatives of any order and number of dimensions
            - The unitary tangent, normal and binormal vectors (Frenet-Serret reference frame) in 2D and 3D
            - The analytic curvature and torsion in 2D and 3D
            - The arc length of the curve in any number of dimensions.
                The arc length is compute by numerical quadrature using analytic derivative information

        The class can be used to represent polynomial and rational Bézier, B-Spline and NURBS curves
        The type of curve depends on the initialization arguments

            - Polymnomial Bézier: Provide the array of control points
            - Rational Bézier:    Provide the arrays of control points and weights
            - B-Spline:           Provide the array of control points, degree and knot vector
            - NURBS:              Provide the arrays of control points and weights, degree and knot vector

        In addition, this class supports operations with real and complex numbers
        The data type used for the computations is detected from the data type of the arguments
        Using complex numbers can be useful to compute the derivative of the shape using the complex step method


        References
        ----------
        The NURBS Book. See references to equations and algorithms throughout the code
        L. Piegl and W. Tiller
        Springer, second edition

        Curves and Surfaces for CADGD. See references to equations the source code
        G. Farin
        Morgan Kaufmann Publishers, fifth edition

        All references correspond to The NURBS book unless it is explicitly stated that they come from Farin's book

    """


    # ---------------------------------------------------------------------------------------------------------------- #
    # Select the type of curve based on the initialization arguments
    # ---------------------------------------------------------------------------------------------------------------- #
    def __init__(self, control_points=None, weights=None, degree=None, knots=None):


        # Void initialization
        if control_points is None and weights is None and degree is None and knots is None:
            pass


        # Polynomial Bezier curve initialization
        elif weights is None and degree is None and knots is None:

            # Set the curve type flag
            self.curve_type = 'Bezier'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]

            # Maximum index of the control points (counting from zero)
            n = np.shape(control_points)[1] - 1

            # Define the weight of the control points
            weights = np.ones((n + 1), dtype=control_points.dtype)

            # Define the order of the basis polynomials
            degree = n

            # Define the knot vector (clamped spline)
            knots = np.concatenate((np.zeros(degree), np.linspace(0, 1, n - degree + 2), np.ones(degree)))


        # Rational Bezier curve initialization
        elif degree is None and knots is None:

            # Set the curve type flag
            self.curve_type = 'R-Bezier'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]

            # Maximum index of the control points (counting from zero)
            n = np.shape(control_points)[1] - 1

            # Define the order of the basis polynomials
            degree = n

            # Define the knot vector (clamped spline)
            knots = np.concatenate((np.zeros(degree), np.linspace(0, 1, n - degree + 2), np.ones(degree)))


        # B-Spline curve initialization (both degree and knot vector are provided)
        elif weights is None and knots is not None:

            # Set the curve type flag
            self.curve_type = 'B-Spline'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]

            # Maximum index of the control points (counting from zero)
            n = np.shape(control_points)[1] - 1

            # Define the weight of the control points
            weights = np.ones((n + 1), dtype=control_points.dtype)


        # B-Spline curve initialization (degree is given but the knot vector is not provided)
        elif weights is None and knots is None:

            # Set the curve type flag
            self.curve_type = 'B-Spline'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]

            # Maximum index of the control points (counting from zero)
            n = np.shape(control_points)[1] - 1

            # Define the knot vector (clamped spline)
            knots = np.concatenate((np.zeros(degree), np.linspace(0, 1, n - degree + 2), np.ones(degree)))

            # Define the weight of the control points
            weights = np.ones((n + 1), dtype=control_points.dtype)


        # NURBS curve initialization
        else:

            # Set the curve type flag
            self.curve_type = 'NURBS'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]


        # Declare input variables as instance variables
        self.P = control_points
        self.W = weights
        self.p = degree
        self.U = knots



    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the coordinates of the curve
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_value(self, u):

        """ Evaluate the coordinates of the curve for the input u parametrization

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Parameter used to evaluate the curve

        Returns
        -------
        C : ndarray with shape (ndim, N)
            Array containing the coordinates of the curve
            The first dimension of ´C´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´C´ spans the ´u´ parametrization sample points

        """

        # Evaluate the NURBS curve for the the input u-parametrization
        C = self.compute_nurbs_coordinates(self.P, self.W, self.p, self.U, u)

        return C


    @staticmethod
    def compute_nurbs_coordinates(P, W, p, U, u):

        """ Evaluate the coordinates of a NURBS curve for the input u parametrization

        This function computes the coordinates of the NURBS curve in homogeneous space using equation 4.5 and then
        maps the coordinates to ordinary space using the perspective map given by equation 1.16. See algorithm A4.1

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points ´(0,1,...,n)´

        W : ndarray with shape (n+1,)
            Array containing the weight of the control points

        p : int
            Degree of the B-Spline basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            Parameter used to evaluate the curve

        Returns
        -------
        C : ndarray with shape (ndim, N)
            Array containing the coordinates of the curve
            The first dimension of ´C´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´C´ spans the ´u´ parametrization sample points

        """

        # Check the shape of the input parameters
        if P.ndim > 2:           raise Exception('P must be an array of shape (ndim, n+1)')
        if W.ndim > 1:           raise Exception('W must be an array of shape (n+1,)')
        if not np.isscalar(p):   raise Exception('p must be an scalar')
        if U.ndim > 1:           raise Exception('U must be an array of shape (r+1=n+p+2,)')
        if np.isscalar(u):       u = np.asarray(u)
        elif u.ndim > 1:         raise Exception('u must be a scalar or an array of shape (N,)')

        # Shape of the array of control points
        n_dim, nn = np.shape(P)

        # Highest index of the control points (counting from zero)
        n = nn - 1

        # Compute the B-Spline basis polynomials
        N_basis = compute_basis_polynomials(n, p, U, u)

        # Map the control points to homogeneous space | P_w = (x*w,y*w,z*w,w)
        P_w = np.concatenate((P * W[np.newaxis, :], W[np.newaxis, :]), axis=0)

        # Compute the coordinates of the NURBS curve in homogeneous space
        # The summations over n is performed exploiting matrix multiplication (vectorized code)
        C_w = np.dot(P_w, N_basis)

        # Map the coordinates back to the ordinary space
        C = C_w[0:-1,:]/C_w[-1, :]

        return C


    @staticmethod
    def compute_bspline_coordinates(P, p, U, u):

        """ Evaluate the coordinates of a B-Spline curve for the input u parametrization

        This function computes the coordinates of a B-Spline curve as given by equation 3.1. See algorithm A3.1

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points ´(0,1,...,n)´

        p : int
            Degree of the B-Spline basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            Parameter used to evaluate the curve

        Returns
        -------
        C : ndarray with shape (ndim, N)
            Array containing the coordinates of the curve
            The first dimension of ´C´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´C´ spans the ´u´ parametrization sample points

        """

        # Check the shape of the input parameters
        if P.ndim > 2:           raise Exception('P must be an array of shape (ndim, n+1)')
        if not np.isscalar(p):   raise Exception('p must be an scalar')
        if U.ndim > 1:           raise Exception('U must be an array of shape (r+1=n+p+2,)')
        if np.isscalar(u):       u = np.asarray(u)
        elif u.ndim > 1:         raise Exception('u must be a scalar or an array of shape (N,)')

        # Highest index of the control points (counting from zero)
        n = np.shape(P)[1] - 1

        # Compute the B-Spline basis polynomials
        N_basis = compute_basis_polynomials(n, p, U, u)

        # Compute the coordinates of the B-Spline curve
        # The summation over n is performed exploiting matrix multiplication (vectorized code)
        C = np.dot(P, N_basis)

        return C


    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the derivatives of the curve
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_derivative(self, u, order):

        """ Evaluate the derivative of the curve for the input u-parametrization

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the curve

        order : integer
            Order of the derivative

        Returns
        -------
        dC : ndarray with shape (ndim, N)
            Array containing the derivative of the desired order
            The first dimension of ´dC´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´dC´ spans the ´u´ parametrization sample points

        """

        # Compute the array of curve derivatives up to order `derivative_order` and slice the desired values
        dC = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.U, u, order)[order, ...]

        return dC


    def compute_nurbs_derivatives(self, P, W, p, U, u, up_to_order):

        """ Compute the derivatives of a NURBS curve in ordinary space up to the desired order

        This function computes the analytic derivatives of the NURBS curve in ordinary space using equation 4.8 and
        the derivatives of the NURBS curve in homogeneous space obtained from compute_bspline_derivatives()

        The derivatives are computed recursively in a fashion similar to algorithm A4.2

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points ´(x,y,z)´
            The second dimension of ´P´ spans the u-direction control points ´(0,1,...,n)´

        W : ndarray with shape (n+1,)
            Array containing the weight of the control points

        p : int
            Degree of the B-Spline basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            Parameter used to evaluate the curve

        up_to_order : integer
            Order of the highest derivative

        Returns
        -------
        nurbs_derivatives: ndarray of shape (up_to_order+1, ndim, Nu)
            The first dimension spans the order of the derivatives (0, 1, 2, ...)
            The second dimension spans the coordinates (x,y,z,...)
            The third dimension spans u-parametrization sample points

        """

        # Set the data type used to initialize arrays (set `complex` if an argument is complex and `float` if not)
        u = np.asarray(u)
        if (P.dtype == np.complex128) or (W.dtype == np.complex128) or (u.dtype == np.complex128):
            data_type = np.complex128
        else:
            data_type = np.float64

        # Map the control points to homogeneous space | P_w = (x*w,y*w,z*w,w)
        P_w = np.concatenate((P * W[np.newaxis, :], W[np.newaxis, :]), axis=0)

        # Compute the derivatives of the NURBS curve in homogeneous space
        bspline_derivatives = self.compute_bspline_derivatives(P_w, p, U, u, up_to_order)
        A_ders = bspline_derivatives[:, 0:-1, :]
        w_ders = bspline_derivatives[:, [-1], :]

        # Initialize array of derivatives
        n_dim, Nu = np.shape(P)[0], np.asarray(u).size                               # Get sizes
        nurbs_derivatives = np.zeros((up_to_order+1, n_dim, Nu), dtype=data_type)    # Initialize array with zeros

        # Compute the derivatives of up to the desired order
        # See algorithm A4.2 from the NURBS book
        for order in range(up_to_order+1):

            # Update the numerator of equation 4.8 recursively
            temp_numerator = A_ders[[order], ...]
            for i in range(1, order+1):
                temp_numerator -= scipy.special.binom(order, i) * w_ders[[i], ...] * nurbs_derivatives[[order-i], ...]

            # Compute the k-th order NURBS curve derivative
            nurbs_derivatives[order, ...] = temp_numerator/w_ders[[0], ...]

        return nurbs_derivatives


    @staticmethod
    def compute_bspline_derivatives(P, p, U, u, up_to_order):

        """ Compute the derivatives of a B-Spline (or NURBS curve in homogeneous space) up to order `derivative_order`

        This function computes the analytic derivatives of a B-Spline curve using equation 3.3. See algorithm A3.2

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points ´(x,y,z)´
            The second dimension of ´P´ spans the u-direction control points ´(0,1,...,n)´

        p : int
            Degree of the B-Spline basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            Parameter used to evaluate the curve

        up_to_order : integer
            Order of the highest derivative

        Returns
        -------
        bspline_derivatives: ndarray of shape (up_to_order+1, ndim, Nu)
            The first dimension spans the order of the derivatives (0, 1, 2, ...)
            The second dimension spans the coordinates (x,y,z,...)
            The third dimension spans u-parametrization sample points

        """

        # Set the data type used to initialize arrays (set `complex` if an argument is complex and `float` if not)
        u = np.asarray(u)
        if (P.dtype == np.complex128) or (u.dtype == np.complex128):
            data_type = np.complex128
        else:
            data_type = np.float64

        # Set the B-Spline coordinates as the zero-th derivative
        n_dim, Nu = np.shape(P)[0], np.asarray(u).size
        bspline_derivatives = np.zeros((up_to_order+1, n_dim, Nu), dtype=data_type)

        # Compute the derivatives of up to the desired order (start at index 1 and end at index `p`)
        # See algorithm A3.2 from the NURBS book
        for order_u in range(min(p, up_to_order) + 1):

            # Highest index of the control points (counting from zero)
            n = np.shape(P)[1] - 1

            # Compute the B-Spline basis polynomials
            N_basis = compute_basis_polynomials_derivatives(n, p, U, u, order_u)

            # Compute the coordinates of the B-Spline
            # The summations over n is performed exploiting matrix multiplication (vectorized code)
            bspline_derivatives[order_u, :, :] = np.dot(P, N_basis)

        # Note that derivative with order higher than `p` are not computed and are be zero from initialization
        # These zero-derivatives are required to compute the higher order derivatives of rational curves

        return bspline_derivatives


    # ---------------------------------------------------------------------------------------------------------------- #
    # Miscellaneous methods
    # ---------------------------------------------------------------------------------------------------------------- #
    def attach_nurbs(self, new_nurbs):

        """ Attatch a new NURBS curve to the end of the instance NURBS curve and return the merged NURBS curve"""

        # Check that the NURBS curves have the same degree
        if self.p != new_nurbs.p:
            raise Exception("In order to merge, the two NURBS curves must have the same degree")

        # Combine the control points
        P = np.concatenate((self.P, new_nurbs.P), axis=1)

        # Combine the control point weights
        W = np.concatenate((self.W, new_nurbs.W), axis=0)

        # Highest index of the control points
        n1 = np.shape(self.P)[1] - 1
        n2 = np.shape(new_nurbs.P)[1] - 1

        # Combine the knot vectors (inner knot has p+1 multiplicity)
        U_start = np.zeros((self.p + 1,))
        U_end = np.ones((self.p + 1,))
        U_mid = np.ones((self.p + 1,)) / 2
        U1 = 0.00 + self.U[self.p + 1:n1 + 1] / 2
        U2 = 0.50 + new_nurbs.U[self.p + 1:n2 + 1] / 2
        U = np.concatenate((U_start, U1, U_mid, U2, U_end))

        # Create the merged NURBS curve
        mergedNurbs = NurbsCurve(control_points=P, weights=W, degree=self.p, knots=U)

        return mergedNurbs


    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the Frenet-Serret unitary vectors
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_tangent(self, u):

        """ Evaluate the unitary tangent vector to the curve for the input u-parametrization

        The definition of the unitary tangent vector is given by equation 10.5 (Farin's textbook)

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the function

        Returns
        -------
        tangent : ndarray with shape (ndim, N)
            Array containing the unitary tangent vector
            The first dimension of ´tangent´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´tangent´ spans the ´u´ parametrization sample points

        """


        # Compute the curve derivatives
        dC, = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.U, u, up_to_order=1)[[1], ...]

        # Compute the tangent vector
        numerator = dC
        denominator = (np.sum(numerator ** 2, axis=0)) ** (1 / 2)
        tangent = numerator / denominator

        return tangent


    def get_normal(self, u):

        """ Evaluate the unitary normal vector to the curve for the input u-parametrization

        The definition of the unitary normal vector is given by equation 10.5 (Farin's textbook)

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the function

        Returns
        -------
        normal : ndarray with shape (ndim, N)
            Array containing the unitary normal vector
            The first dimension of ´normal´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´normal´ spans the ´u´ parametrization sample points

        """

        # Compute the curve derivatives
        dC, ddC = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.U, u, up_to_order=2)[[1, 2], ...]


        # Compute the normal vector
        if self.ndim == 2:
            dC = np.concatenate((dC, np.zeros((1, np.asarray(u).size))), axis=0)
            ddC = np.concatenate((ddC, np.zeros((1, np.asarray(u).size))), axis=0)
            numerator = np.cross(dC, np.cross(ddC, dC, axisa=0, axisb=0, axisc=0), axisa=0, axisb=0, axisc=0)
            denominator = (np.sum(numerator ** 2, axis=0)) ** (1 / 2)
            normal = (numerator / denominator)[[0, 1], ...]

        elif self.ndim == 3:
            numerator = np.cross(dC, np.cross(ddC, dC, axisa=0, axisb=0, axisc=0), axisa=0, axisb=0, axisc=0)
            denominator = (np.sum(numerator ** 2, axis=0)) ** (1 / 2)
            normal = numerator / denominator

        else: raise Exception("The number of dimensions must be 2 or 3")

        return normal


    def get_binormal(self, u):

        """ Evaluate the unitary binormal vector to the curve for the input u-parametrization

        The definition of the unitary binormal vector is given by equation 10.5 (Farin's textbook)

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the function

        Returns
        -------
        binormal : ndarray with shape (ndim, N)
            Array containing the unitary binormal vector
            The first dimension of ´binormal´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´binormal´ spans the ´u´ parametrization sample points

        """

        # Compute the curve derivatives
        dC, ddC = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.U, u, up_to_order=2)[[1, 2], ...]

        # Compute the binormal vector
        if self.ndim == 2:
            dC = np.concatenate((dC, np.zeros((1, np.asarray(u).size))), axis=0)
            ddC = np.concatenate((ddC, np.zeros((1, np.asarray(u).size))), axis=0)
            numerator = np.cross(dC, ddC, axisa=0, axisb=0, axisc=0)
            denominator = (np.sum(numerator ** 2, axis=0)) ** (1 / 2)
            binormal = (numerator / denominator)[[0, 1], ...]

        elif self.ndim == 3:
            numerator = np.cross(dC, ddC, axisa=0, axisb=0, axisc=0)
            denominator = (np.sum(numerator ** 2, axis=0)) ** (1 / 2)
            binormal = numerator / denominator

        else: raise Exception("The number of dimensions must be 2 or 3")

        return binormal


    def get_normal_2D(self, u):

        """ Evaluate the the unitary normal vector using the special formula 2D formula

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the function

        Returns
        -------
        normal : ndarray with shape (2, N)
            Array containing the unitary normal vector
            The first dimension of ´normal´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´normal´ spans the ´u´ parametrization sample points

        """

        # Compute the curve derivatives
        u = np.asarray(u)
        dC = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.U, u, up_to_order=1)[1, ...]

        # Compute the normal vector
        if self.ndim == 2:
            numerator = np.concatenate((-dC[[1], :], dC[[0], :]), axis=0)
            denominator = (np.sum(numerator ** 2, axis=0)) ** (1 / 2)
            normal = numerator / denominator

        else: raise Exception("The number of dimensions must be 2")

        return normal


    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the curvature, torsion, and arc-length
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_curvature(self, u):

        """ Evaluate the curvature of the curve for the input u-parametrization

        The definition of the curvature is given by equation 10.7 (Farin's textbook)

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the curvature

        Returns
        -------
        curvature : scalar or ndarray with shape (N, )
            Scalar or array containing the curvature of the curve

        """

        # Compute the curve derivatives
        u = np.asarray(u)
        dC, ddC = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.U, u, up_to_order=2)[[1, 2], ...]

        # Compute the curvature
        if self.ndim == 2:
            dC = np.concatenate((dC, np.zeros((1, np.asarray(u).size))), axis=0)
            ddC = np.concatenate((ddC, np.zeros((1, np.asarray(u).size))), axis=0)
            numerator = np.sum(np.cross(ddC, dC, axisa=0, axisb=0, axisc=0) ** 2, axis=0) ** (1 / 2)
            denominator = (np.sum(dC ** 2, axis=0)) ** (3 / 2)
            curvature = (numerator / denominator)

        elif self.ndim == 3:
            numerator = np.sum(np.cross(ddC, dC, axisa=0, axisb=0, axisc=0) ** 2, axis=0) ** (1 / 2)
            denominator = (np.sum(dC ** 2, axis=0)) ** (3 / 2)
            curvature = numerator / denominator

        else: raise Exception("The number of dimensions must be 2 or 3")

        return curvature


    def get_torsion(self, u):

        """ Evaluate the torsion of the curve for the input u-parametrization

        The definition of the torsion is given by equation 10.8 (Farin's textbook)

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the torsion

        Returns
        -------
        torsion : scalar or ndarray with shape (N, )
            Scalar or array containing the torsion of the curve

        """

        # Compute the curve derivatives
        u = np.asarray(u)
        dC, ddC, dddC = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.U, u, up_to_order=3)[[1, 2, 3], ...]

        # Compute the torsion
        if self.ndim == 2:
            dC = np.concatenate((dC, np.zeros((1, np.asarray(u).size))), axis=0)
            ddC = np.concatenate((ddC, np.zeros((1, np.asarray(u).size))), axis=0)
            dddC = np.concatenate((dddC, np.zeros((1, np.asarray(u).size))), axis=0)
            numerator = np.sum(np.cross(dC, ddC, axisa=0, axisb=0, axisc=0) * dddC, axis=0)
            denominator = np.sum(np.cross(dC, ddC, axisa=0, axisb=0, axisc=0)**2, axis=0)
            torsion = (numerator / denominator)

        elif self.ndim == 3:
            numerator = np.sum(np.cross(dC, ddC, axisa=0, axisb=0, axisc=0) * dddC, axis=0)
            denominator = np.sum(np.cross(dC, ddC, axisa=0, axisb=0, axisc=0)**2, axis=0)
            torsion = numerator / denominator

        else: raise Exception("The number of dimensions must be 2 or 3")

        return torsion


    def get_arclength(self, u1=0.00, u2=1.00):

        """ Compute the arc length of a parametric curve in the interval [u1,u2] using numerical quadrature

        The definition of the arc length is given by equation 10.3 (Farin's textbook)

        Parameters
        ----------
        u1 : scalar
            Lower limit of integration for the arc length computation

        u2 : scalar
            Upper limit of integration for the arc length computation

        Returns
        -------
        L : scalar
            Arc length of NURBS curve in the interval [u1, u2]

        """

        # Compute the arc length differential using the central finite differences
        def get_arclegth_differential(u):
            dCdu = self.get_derivative(u, order=1)
            dLdu = np.sqrt(np.sum(dCdu ** 2, axis=0))  # dL/du = [(dx_0/du)^2 + ... + (dx_n/du)^2]^(1/2)
            return dLdu

        # Compute the arc length of C(t) in the interval [u1, u2] by numerical integration
        arclength = scipy.integrate.fixed_quad(get_arclegth_differential, u1, u2, n=10)[0]

        return arclength



    # ---------------------------------------------------------------------------------------------------------------- #
    # Plot the NURBS curve
    # ---------------------------------------------------------------------------------------------------------------- #
    def plot(self, fig=None, ax=None,
             curve=True, control_points=True, frenet_serret=False, axis_off=False, ticks_off=False):

        """ Create a plot and return the figure and axes handles """

        if fig is None:
            # One dimension (law of evolution)
            if self.ndim == 1:
                fig = plt.figure(figsize=(6, 5))
                ax = fig.add_subplot(111)
                ax.set_xlabel('$u$ parameter', fontsize=12, color='k', labelpad=12)
                ax.set_ylabel('NURBS curve value', fontsize=12, color='k', labelpad=12)
                # ax_xy.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
                # ax_xy.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
                for t in ax.xaxis.get_major_ticks(): t.label.set_fontsize(12)
                for t in ax.yaxis.get_major_ticks(): t.label.set_fontsize(12)
                if ticks_off:
                    ax.set_xticks([])
                    ax.set_yticks([])
                if axis_off:
                    ax.axis('off')

            # Two dimensions (plane curve)
            elif self.ndim == 2:
                fig = plt.figure(figsize=(6, 5))
                ax = fig.add_subplot(111)
                ax.set_xlabel('$x$ axis', fontsize=12, color='k', labelpad=12)
                ax.set_ylabel('$y$ axis', fontsize=12, color='k', labelpad=12)
                # ax_xy.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
                # ax_xy.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
                for t in ax.xaxis.get_major_ticks(): t.label.set_fontsize(12)
                for t in ax.yaxis.get_major_ticks(): t.label.set_fontsize(12)
                if ticks_off:
                    ax.set_xticks([])
                    ax.set_yticks([])
                if axis_off:
                    ax.axis('off')

            # Three dimensions (space curve)
            elif self.ndim == 3:
                fig = mpl.pyplot.figure(figsize=(6, 5))
                ax = fig.add_subplot(111, projection='3d')
                ax.view_init(azim=-120, elev=30)
                ax.grid(False)
                ax.xaxis.pane.fill = False
                ax.yaxis.pane.fill = False
                ax.zaxis.pane.fill = False
                ax.xaxis.pane.set_edgecolor('k')
                ax.yaxis.pane.set_edgecolor('k')
                ax.zaxis.pane.set_edgecolor('k')
                ax.xaxis.pane._alpha = 0.9
                ax.yaxis.pane._alpha = 0.9
                ax.zaxis.pane._alpha = 0.9
                ax.set_xlabel('$x$ axis', fontsize=12, color='k', labelpad=12)
                ax.set_ylabel('$y$ axis', fontsize=12, color='k', labelpad=12)
                ax.set_zlabel('$z$ axis', fontsize=12, color='k', labelpad=12)
                # ax_xy.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
                # ax_xy.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
                # ax_xy.zaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
                for t in ax.xaxis.get_major_ticks(): t.label.set_fontsize(8)
                for t in ax.yaxis.get_major_ticks(): t.label.set_fontsize(8)
                for t in ax.zaxis.get_major_ticks(): t.label.set_fontsize(8)
                ax.xaxis.set_rotate_label(False)
                ax.yaxis.set_rotate_label(False)
                ax.zaxis.set_rotate_label(False)
                if ticks_off:
                    ax.set_xticks([])
                    ax.set_yticks([])
                    ax.set_zticks([])
                if axis_off:
                    ax.axis('off')

            else: raise Exception('The number of dimensions must be 1, 2 or 3')


        # Add objects to the plot
        if curve:          self.plot_curve(fig, ax)
        if control_points: self.plot_control_points(fig, ax)
        if frenet_serret:  self.plot_frenet_serret(fig, ax)

        # Set the scaling of the axes
        self.rescale_plot(fig, ax)

        return fig, ax


    def plot_curve(self, fig, ax, linewidth=1.5, linestyle='-', color='black'):

        """ Plot the coordinates of the NURBS curve """

        # One dimension (law of evolution)
        if self.ndim == 1:
            u = np.linspace(0, 1, 501)
            X = np.real(self.get_value(u))
            line, = ax.plot(u, X[0,:])
            line.set_linewidth(linewidth)
            line.set_linestyle(linestyle)
            line.set_color(color)
            line.set_marker(' ')
            # line.set_label(' ')

        # Two dimensions (plane curve)
        elif self.ndim == 2:
            u = np.linspace(0, 1.00, 501)
            X, Y = np.real(self.get_value(u))
            line, = ax.plot(X, Y)
            line.set_linewidth(linewidth)
            line.set_linestyle(linestyle)
            line.set_color(color)
            line.set_marker(' ')
            # line.set_label(' ')

        # Three dimensions (space curve)
        elif self.ndim == 3:
            u = np.linspace(0, 1, 501)
            X, Y, Z = np.real(self.get_value(u))
            line, = ax.plot(X, Y, Z)
            line.set_linewidth(linewidth)
            line.set_linestyle(linestyle)
            line.set_color(color)
            line.set_marker(' ')
            # line.set_label(' ')

        else: raise Exception('The number of dimensions must be 1, 2 or 3')

        return fig, ax


    def plot_control_points(self, fig, ax, linewidth=1.25, linestyle='-.', color='red', markersize=5, markerstyle='o'):

        """ Plot the control points of the NURBS curve """

        # One dimension (law of evolution)
        if self.ndim == 1:
            Px = np.real(self.P)
            u = np.linspace(0, 1, Px.size)
            line, = ax.plot(u, Px[0,:])
            line.set_linewidth(linewidth)
            line.set_linestyle(linestyle)
            line.set_color(color)
            line.set_marker(markerstyle)
            line.set_markersize(markersize)
            line.set_markeredgewidth(linewidth)
            line.set_markeredgecolor(color)
            line.set_markerfacecolor('w')
            line.set_zorder(4)
            # line.set_label(' ')


        # Two dimensions (plane curve)
        elif self.ndim == 2:
            Px, Py = np.real(self.P)
            line, = ax.plot(Px, Py)
            line.set_linewidth(linewidth)
            line.set_linestyle(linestyle)
            line.set_color(color)
            line.set_marker(markerstyle)
            line.set_markersize(markersize)
            line.set_markeredgewidth(linewidth)
            line.set_markeredgecolor(color)
            line.set_markerfacecolor('w')
            line.set_zorder(4)
            # line.set_label(' ')

        # Three dimensions (space curve)
        elif self.ndim == 3:
            Px, Py, Pz = np.real(self.P)
            line, = ax.plot(Px, Py, Pz)
            line.set_linewidth(linewidth)
            line.set_linestyle(linestyle)
            line.set_color(color)
            line.set_marker(markerstyle)
            line.set_markersize(markersize)
            line.set_markeredgewidth(linewidth)
            line.set_markeredgecolor(color)
            line.set_markerfacecolor('w')
            line.set_zorder(4)
            # line.set_label(' ')

        else: raise Exception('The number of dimensions must be 2 or 3')

        return fig, ax


    def plot_frenet_serret(self, fig, ax, frame_number=5, frame_scale=0.10):

        """ Plot some Frenet-Serret reference frames along the NURBS curve """

        # Compute the tangent, normal and binormal unitary vectors
        h = 1e-12
        u = np.linspace(0+h, 1-h, frame_number)
        position = np.real(self.get_value(u))
        tangent = np.real(self.get_tangent(u))
        normal = np.real(self.get_normal(u))
        binormal = np.real(self.get_binormal(u))

        # Two dimensions (plane curve)
        if self.ndim == 2:

            # Plot the frames of reference
            for k in range(frame_number):

                # Plot the tangent vector
                x, y = position[:, k]
                u, v = tangent[:, k]
                ax.quiver(x, y, u, v, color='red', scale=7.5)

                # Plot the normal vector
                x, y = position[:, k]
                u, v = normal[:, k]
                ax.quiver(x, y, u, v, color='blue', scale=7.5)

            # Plot the origin of the vectors
            x, y = position
            points, = ax.plot(x, y)
            points.set_linestyle(' ')
            points.set_marker('o')
            points.set_markersize(5)
            points.set_markeredgewidth(1.25)
            points.set_markeredgecolor('k')
            points.set_markerfacecolor('w')
            points.set_zorder(4)
            # points.set_label(' ')

        # Three dimensions (space curve)
        elif self.ndim == 3:

            # Compute a length scale (fraction of the curve arc length)
            scale = frame_scale * self.get_arclength(0, 1)

            # Plot the frames of reference
            for k in range(frame_number):

                # Plot the tangent vector
                x, y, z = position[:, k]
                u, v, w = tangent[:, k]
                ax.quiver(x, y, z, u, v, w, color='red', length=scale, normalize=True)

                # Plot the norma vector
                x, y, z = position[:, k]
                u, v, w = normal[:, k]
                ax.quiver(x, y, z, u, v, w, color='blue', length=scale, normalize=True)

                # Plot the binormal vector
                x, y, z = position[:, k]
                u, v, w = binormal[:, k]
                ax.quiver(x, y, z, u, v, w, color='green', length=scale, normalize=True)

            # Plot the origin of the vectors
            x, y, z = position
            points, = ax.plot(x, y, z)
            points.set_linestyle(' ')
            points.set_marker('o')
            points.set_markersize(5)
            points.set_markeredgewidth(1.25)
            points.set_markeredgecolor('k')
            points.set_markerfacecolor('w')
            points.set_zorder(4)
            # points.set_label(' ')


        else: raise Exception('The number of dimensions must be 2 or 3')

        return fig, ax


    def rescale_plot(self, fig, ax):

        """ Adjust the aspect ratio of the figure """

        # Two dimensions (plane curve)
        if self.ndim == 2:

            # Set the aspect ratio of the data
            ax.set_aspect(1.0)

            # Adjust pad
            plt.tight_layout(pad=5.0, w_pad=None, h_pad=None)

        # Three dimensions (space curve)
        if self.ndim == 3:

            # Set axes aspect ratio
            ax.autoscale(enable=True)
            x_min, x_max = ax.get_xlim()
            y_min, y_max = ax.get_ylim()
            z_min, z_max = ax.get_zlim()
            x_mid = (x_min + x_max) / 2
            y_mid = (y_min + y_max) / 2
            z_mid = (z_min + z_max) / 2
            L = np.max((x_max - x_min, y_max - y_min, z_max - z_min)) / 2
            ax.set_xlim3d(x_mid - 1.0 * L, x_mid + 1.0 * L)
            ax.set_ylim3d(y_mid - 1.0 * L, y_mid + 1.0 * L)
            ax.set_zlim3d(z_mid - 1.0 * L, z_mid + 1.0 * L)

            # Adjust pad
            plt.tight_layout(pad=5.0, w_pad=None, h_pad=None)


    def plot_curvature(self, fig=None, ax=None, color='black', linestyle='-'):

        # Create the figure
        if fig is None:
            fig = plt.figure(figsize=(6, 5))
            ax = fig.add_subplot(111)
        ax.set_xlabel('$u$ parameter', fontsize=12, color='k', labelpad=12)
        ax.set_ylabel('Curvature', fontsize=12, color='k', labelpad=12)
        # ax_xy.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
        # ax_xy.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
        for t in ax.xaxis.get_major_ticks(): t.label.set_fontsize(12)
        for t in ax.yaxis.get_major_ticks(): t.label.set_fontsize(12)
        # ax_xy.set_xticks([])
        # ax_xy.set_yticks([])
        # ax_xy.axis('off')

        # Plot the curvature distribution
        u = np.linspace(0, 1, 1000)
        curvature = np.real(self.get_curvature(u))
        line, = ax.plot(u, curvature)
        line.set_linewidth(1.25)
        line.set_linestyle(linestyle)
        line.set_color(color)
        line.set_marker(" ")
        line.set_markersize(3.5)
        line.set_markeredgewidth(1)
        line.set_markeredgecolor("k")
        line.set_markerfacecolor("w")
        # line.set_label(' ')

        # Set the aspect ratio of the data
        # ax_xy.set_aspect(1.0)

        # # Set the aspect ratio of the figure
        # ratio = 1.00
        # x1, x2 = ax_xy.get_xlim()
        # y1, y2 = ax_xy.get_ylim()
        # ax_xy.set_aspect(np.abs((x2-x1)/(y2-y1))*ratio)

        # Adjust pad
        plt.tight_layout(pad=5.0, w_pad=None, h_pad=None)

        return fig, ax

    def plot_torsion(self, fig=None, ax=None, color='black', linestyle='-'):

        # Create the figure
        if fig is None:
            fig = plt.figure(figsize=(6, 5))
            ax = fig.add_subplot(111)
        ax.set_xlabel('$u$ parameter', fontsize=12, color='k', labelpad=12)
        ax.set_ylabel('Torsion', fontsize=12, color='k', labelpad=12)
        # ax_xy.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
        # ax_xy.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
        for t in ax.xaxis.get_major_ticks(): t.label.set_fontsize(12)
        for t in ax.yaxis.get_major_ticks(): t.label.set_fontsize(12)
        # ax_xy.set_xticks([])
        # ax_xy.set_yticks([])
        # ax_xy.axis('off')

        # Plot the curvature distribution
        u = np.linspace(0, 1, 1000)
        torsion = np.real(self.get_torsion(u))
        line, = ax.plot(u, torsion)
        line.set_linewidth(1.25)
        line.set_linestyle(linestyle)
        line.set_color(color)
        line.set_marker(" ")
        line.set_markersize(3.5)
        line.set_markeredgewidth(1)
        line.set_markeredgecolor("k")
        line.set_markerfacecolor("w")
        # line.set_label(' ')

        # Set the aspect ratio of the data
        # ax_xy.set_aspect(1.0)

        # Set the aspect ratio of the figure
        ratio = 1.00
        x1, x2 = ax.get_xlim()
        y1, y2 = ax.get_ylim()
        ax.set_aspect(np.abs((x2 - x1) / (y2 - y1)) * ratio)

        # Adjust pad
        plt.tight_layout(pad=5.0, w_pad=None, h_pad=None)

        return fig, ax




# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np

from .nurbs_surface import NurbsSurface


# -------------------------------------------------------------------------------------------------------------------- #
# Define the bilinear NURBS surface class
# -------------------------------------------------------------------------------------------------------------------- #
class NurbsSurfaceBilinear:

    """ Create a NURBS representation of the bilinear patch defined by corners P00, P01, P10, and P11

        Create a NURBS representation of the bilinear patch
            S(u,v) = (1-v)*[(1-u)*P00 + u*P01] + v*[(1-u)*P10 + u*P11]

        Note that a bilinear patch is a ruled surface with segments (P00, P01) and (P10, P11) as generating curves
            S(u,v) = (1-v)*C1(u) + v*C2(u)
            C1(u) = (1-u)*P00 + u*P01
            C2(u) = (1-u)*P10 + u*P11

    Parameters
    ----------
    P00, P01, P10, P11 : ndarrays with shape (ndim,)
        Coordinates of the corner points defining the bilinear surface (ndim=3)

    References
    ----------
    The NURBS book. Chapter 8.2
    L. Piegl and W. Tiller
    Springer, second edition

    """

    def __init__(self, P00, P01, P10, P11):

        # Declare input variables as instance variables
        self.P00 = P00
        self.P01 = P01
        self.P10 = P10
        self.P11 = P11
        self.ndim = 3

        # Check the number of dimensions of the problem
        ndims = [np.shape(P00)[0], np.shape(P01)[0], np.shape(P10)[0], np.shape(P11)[0]]
        if any([ndim != 3 for ndim in ndims]):
            raise Exception("The input points must be three-dimensional")

        # Make the bilinear patch NURBS representation
        self.NurbsSurface = None
        self.make_nurbs_surface()


    def make_nurbs_surface(self):

        """ Make a NURBS surface representation of the bilinear surface """

        # Define the array of control points
        n_dim, n, m = self.ndim, 2, 2
        P = np.zeros((n_dim, n, m))
        P[:, 0, 0] = self.P00
        P[:, 1, 0] = self.P01
        P[:, 0, 1] = self.P10
        P[:, 1, 1] = self.P11

        # Create the NURBS surface
        self.NurbsSurface = NurbsSurface(control_points=P)






# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np

from .nurbs_surface import NurbsSurface


# -------------------------------------------------------------------------------------------------------------------- #
# Define the bilinear NURBS surface class
# -------------------------------------------------------------------------------------------------------------------- #
class NurbsSurfaceRevolution:

    """ Create the NURBS surface obtained by revolving a generatrix NURBS curve `C(u)` about an axis

    Parameters
    ----------
    generatrix : NURBS curve object
        See NurbsSurface class documentation

    axis_point : ndaray of with shape (3,)
        Point that together with a direction defines the axis of rotation

    axis_direction : ndarrays with shape (3,)
        Direction that together with a point defines the axis of rotation

    theta_start : scalar
        Start angle, measured with respect to the generatrix

    theta_end : scalar
        End angle, measured with respect to the generatrix

    References
    ----------
    The NURBS book. Chapter 8.5
    L. Piegl and W. Tiller
    Springer, second edition

    """

    def __init__(self, generatrix, axis_point, axis_direction, angle_start, angle_end):

        # Set the data type used to initialize arrays (set `complex` if any argument is complex and `float` if not)
        temp = generatrix.P
        for item in locals().values():
            data_type = np.asarray(item).dtype
            if np.issubdtype(data_type, np.complex128):
                self.data_type = np.complex128
                break
            else:
                self.data_type = np.float64

        # Declare input variables as instance variables (adopt the notation used in the NURBS book)
        self.C = generatrix
        self.S = axis_point
        self.T = axis_direction/np.sum(axis_direction**2)**(1/2)
        self.theta_start = angle_start
        self.theta_end = angle_end

        # Check the number of dimensions of the problem
        if np.shape(generatrix.P)[0] != 3:   raise Exception("The input NURBS must be three-dimensional")
        if np.shape(axis_direction)[0] != 3: raise Exception("The axis direction must be a three-dimensional vector")
        if np.shape(axis_point)[0] != 3:     raise Exception("The axis point must be a three-dimensional vector")

        # Make the extrusion surface NURBS representation
        self.NurbsSurface = None
        self.make_nurbs_surface()


    def make_nurbs_surface(self):

        """ Make a NURBS surface representation of the revolution surface """

        # Rename variables for brevity
        S = self.S
        T = self.T
        theta_start = self.theta_start
        theta_end = self.theta_end

        # Correct theta_end if necessary
        if theta_end < theta_start: theta_end = theta_end + 2*np.pi

        # Angle spanned by the circular arc
        theta = theta_end - theta_start

        # Get the number of NURBS segments used to represent the circular arc
        if    theta <= 1/2*np.pi: n_arcs = 1
        elif  theta <= 2/2*np.pi: n_arcs = 2
        elif  theta <= 3/2*np.pi: n_arcs = 3
        elif  theta <= 4/2*np.pi: n_arcs = 4
        else: raise Exception('Opps, something went wrong...')

        # Angle spanned by each segment
        delta_theta = theta/n_arcs

        # Initialize arrays of control points and weights
        n = 2*n_arcs                                                  # Highest index of the u-direction control points
        m = np.shape(self.C.P)[1] - 1                                 # Highest index of the v_direction control points
        P_array = np.zeros((3, n + 1, m + 1), dtype=self.data_type)   # Array of control points
        W_array = np.zeros((n + 1, m + 1), dtype=self.data_type)      # Weight of the control points

        # Loop over the generatrix control points
        for j in range(m+1):

            # Current control point of the generatrix
            P = self.C.P[:, j]

            # Current weight of the generatrix
            W = self.C.W[j]

            # Compute the axis point that is closest to the control point
            O = self.project_point_to_line(S, T, P)

            # Compute the projected distance between the current control point and the axis
            R = np.sum((P-O)**2)**(1/2)

            # Compute the current section principal directions (use dummy directions in case P and O coincide)
            if np.abs(np.sum(P-O)) < 1e-12:
                X = np.asarray([1, 0, 0])
                Y = np.asarray([0, 1, 0])
            else:
                X = (P-O)/np.sum((P-O)**2)**(1/2)
                Y = np.cross(T, X)

            # Get the coordinates and weight of the first control point
            P0 = O + R * np.cos(theta_start) * X + R * np.sin(theta_start) * Y
            T0 = -np.sin(theta_start) * X + np.cos(theta_start) * Y

            # Store control points and weights
            P_array[:, 0, j] = P0
            W_array[0, j] = W

            # Get the coordinates and weights of the other control points
            index, angle = 0, theta_start
            for i in range(n_arcs):

                # Angle spanned by the current segment
                angle = angle + delta_theta

                # Get the end-point and end-tangent of the current segment
                P2 = O + R * np.cos(angle) * X + R * np.sin(angle) * Y
                T2 = -np.sin(angle) * X + np.cos(angle) * Y

                # Solve the intersection between tangent lines T0 and T2 to compute the intermediate point
                P1 = self.intersect_lines(P0, T0, P2, T2)

                # Compute the weight of the intermediate point
                W1 = np.cos(delta_theta / 2)

                # Store control points and weights
                P_array[:, index + 1, j] = P1
                P_array[:, index + 2, j] = P2
                W_array[index + 1, j] = W*W1
                W_array[index + 2, j] = W

                # Get ready for the next segment!
                index = index + 2
                P0 = P2
                T0 = T2

        # Define the order of the basis polynomials
        p = 2
        q = self.C.p

        # Define the knot vectors
        # Knot multiplicity p+1 at the endpoints
        U = 5 + np.zeros((n + p + 2))
        U[[0, 1, 2]] = 0
        U[[-1, -2, -3]] = 1

        # Set the multiplicity 2 at the interior knots to connect the segments
        if n_arcs == 1:
            pass
        elif n_arcs == 2:
            U[[3, 4]] = 1 / 2
        elif n_arcs == 3:
            U[[3, 4]], U[[5, 6]] = 1 / 3, 2 / 3
        elif n_arcs == 4:
            U[[3, 4]], U[[5, 6]], U[[7, 8]] = 1 / 4, 2 / 4, 3 / 4
        else:
            raise Exception('Opps, something went wrong...')

        # The knot vector in the v-direction is given by the knot vector of the generatrix NURBS
        V = self.C.U

        # Create the NURBS surface
        self.NurbsSurface = NurbsSurface(control_points=P_array, weights=W_array, u_degree=p, v_degree=q, u_knots=U, v_knots=V)


    def intersect_lines(self, P0, T0, P2, T2):

        """ Compute the point of intersection between two lines in 2D or 3D """

        # Compute the intersection by reducing the 3x2 system to a 2x2 system using dot products
        A = np.asarray([[np.sum(T0 * T0), -np.sum(T2 * T0)], [np.sum(T0 * T2), -np.sum(T2 * T2)]])
        b = np.asarray([np.sum(P2 * T0) - np.sum(P0 * T0), np.sum(P2 * T2) - np.sum(P0 * T2)])
        u, v = np.linalg.solve(A, b)
        P1 = P0 + u * T0

        if np.sum(np.abs(((P0 + u * T0) - (P2 + v * T2)))) > 1e-12:
            raise Exception("Something went wrong computing the line intersection")

        return P1


    def project_point_to_line(self, S, T, P):

        """ Compute the projection of a point ´P´ into the line given by ´S + u*T´ """

        # Analytic formula (not hard to derive by hand)
        P_projected = S + np.sum(T * (P - S)) / np.sum(T * T) * T

        return P_projected


# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np
import numba as nb


# -------------------------------------------------------------------------------------------------------------------- #
# Compute the basis polynomials
# -------------------------------------------------------------------------------------------------------------------- #
@nb.jit(nopython=True, cache=True)
def compute_basis_polynomials(n, p, U, u, return_degree=None):

    """ Evaluate the n-th B-Spline basis polynomials of degree ´p´ for the input u-parametrization

    The basis polynomials are computed from their definition by implementing equation 2.5 directly

    Parameters
    ----------
    n : integer
        Highest index of the basis polynomials (n+1 basis polynomials)

    p : integer
        Degree of the basis polynomials

    U : ndarray with shape (r+1=n+p+2,)
        Knot vector of the basis polynomials
        Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

    u : scalar or ndarray with shape (Nu,)
        Parameter used to evaluate the basis polynomials (real or complex!)

    return_degree : int
        Degree of the returned basis polynomials

    Returns
    -------
    N : ndarray with shape (n+1, Nu)
        Array containing the basis polynomials of order ´p´ evaluated at ´u´
        The first dimension of ´N´ spans the n-th polynomials
        The second dimension of ´N´ spans the ´u´ parametrization sample points

    """

    # # Check the degree of the basis basis polynomials (not possible to compile with Numba yet)
    # if p < 0: raise Exception('The degree of the basis polynomials cannot be negative')
    #
    # # Check the number of basis basis polynomials (not possible to compile with Numba yet)
    # if p > n: raise Exception('The degree of the basis polynomials must be equal or lower than the number of basis polynomials')

    # Number of points where the polynomials are evaluated (vectorized computations)
    u = np.asarray(u)
    Nu = np.asarray(u).size

    # Number of basis polynomials at the current step of the recursion
    m = n + p + 1

    # Initialize the array of basis polynomials
    N = np.zeros((p + 1, m, Nu), dtype=u.dtype)

    # First step of the recursion formula (p = 0)
    # The case when point_index==n and u==1 is an special case. See the NURBS book section 2.5 and algorithm A2.1
    # The np.real() operator is used such that the function extends to complex u-parameter as well
    for i in range(m):
        # N[0, i, :] = 0.0 + 1.0 * (u >= U[i]) * (u < U[i + 1])
        # N[0, i, :] = 0.0 + 1.0 * (u >= U[i]) * (u < U[i + 1]) + 1.00 * (np.logical_and(u == 1, i == n))
        N[0, i, :] = 0.0 + 1.0 * (np.real(u) >= U[i]) * (np.real(u) < U[i + 1]) + 1.00 * (np.logical_and(np.real(u) == U[-1], i == n))

    # Second and next steps of the recursion formula (p = 1, 2, ...)
    for k in range(1, p + 1):

        # Update the number of basis polynomials
        m = m - 1

        # Compute the basis polynomials using the de Boor recursion formula
        for i in range(m):

            # Compute first factor (avoid division by zero by convention)
            if (U[i + k] - U[i]) == 0:
                n1 = np.zeros((Nu,), dtype=u.dtype)
            else:
                n1 = (u - U[i]) / (U[i + k] - U[i]) * N[k - 1, i, :]

            # Compute second factor (avoid division by zero by convention)
            if (U[i + k + 1] - U[i + 1]) == 0:
                n2 = np.zeros((Nu,), dtype=u.dtype)
            else:
                n2 = (U[i + k + 1] - u) / (U[i + k + 1] - U[i + 1]) * N[k - 1, i + 1, :]

            # Compute basis polynomial (recursion formula 2.5)
            N[k, i, ...] = n1 + n2

    # Get the n+1 basis polynomials of the desired degree
    N = N[p, 0:n+1, :] if return_degree is None else N[return_degree, 0:n+1, :]


    # # Numba assertions
    # assert p >= 0    # The degree of the basis polynomials cannot be negative
    # assert p <= n    # The degree of the basis polynomials must be equal or lower than the number of basis polynomials

    return N



# -------------------------------------------------------------------------------------------------------------------- #
# Compute the derivatives of the basis polynomials
# -------------------------------------------------------------------------------------------------------------------- #
@nb.jit(nopython=True, cache=True)
def compute_basis_polynomials_derivatives(n, p, U, u, derivative_order):

    """ Evaluate the derivative of the n-th B-Spline basis polynomials of degree ´p´ for the input u-parametrization

    The basis polynomials derivatives are computed recursively by implementing equations 2.7 and 2.9 directly

    Parameters
    ----------
    n : integer
        Highest index of the basis polynomials (n+1 basis polynomials)

    p : integer
        Degree of the original basis polynomials

    U : ndarray with shape (r+1=n+p+2,)
        Knot vector of the basis polynomials
        Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

    u : scalar or ndarray with shape (Nu,)
        Parameter used to evaluate the basis polynomials

    derivative_order : scalar
        Order of the basis polynomial derivatives

    Returns
    -------
    N_ders : ndarray with shape (n+1, Nu)
        Array containing the basis spline polynomials derivatives evaluated at ´u´
        The first dimension of ´N´ spans the n-th polynomials
        The second dimension of ´N´ spans the ´u´ parametrization sample points

    """

    # Check the number of derivative order (not possible to compile with Numba yet)
    if derivative_order > p: raise Exception('The derivative order is higher than the degree of the basis polynomials')

    # Compute the basis polynomials to the right hand side of equation 2.9 recursively down to the zeroth derivative
    # Each new call reduces the degree of the basis polynomials by one
    if derivative_order >= 1:
        derivative_order -= 1
        N = compute_basis_polynomials_derivatives(n, p - 1, U, u, derivative_order)
    elif derivative_order == 0:
        N = compute_basis_polynomials(n, p, U, u)
        return N
    else:
        print('Oooopps, something went wrong in compute_basis_polynomials_derivatives()')
        N = compute_basis_polynomials(n, p, U, u)
        return N

    # Number of points where the polynomials are evaluated (vectorized computations)
    Nu = np.asarray(u).size

    # Initialize the array of basis polynomial derivatives
    N_ders = np.zeros((n + 1, Nu), dtype=u.dtype)

    # Compute the derivatives of the (0, 1, ..., n) basis polynomials using equations 2.7 and 2.9
    for i in range(n + 1):

        # Compute first factor (avoid division by zero by convention)
        if (U[i + p] - U[i]) == 0:
            n1 = np.zeros(Nu, dtype=u.dtype)
        else:
            n1 = p * N[i, :] / (U[i + p] - U[i])

        # Compute second factor (avoid division by zero by convention)
        if (U[i + p + 1] - U[i + 1]) == 0:
            n2 = np.zeros(Nu, dtype=u.dtype)
        else:
            n2 = p * N[i + 1, :] / (U[i + p + 1] - U[i + 1])

        # Compute the derivative of the current basis polynomials
        N_ders[i, :] = n1 - n2

    return N_ders



# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np
import scipy.special
import matplotlib as mpl
import matplotlib.pyplot as plt

from mpl_toolkits  import mplot3d
from .nurbs_curve  import NurbsCurve
from .nurbs_basis_functions  import compute_basis_polynomials, compute_basis_polynomials_derivatives


# -------------------------------------------------------------------------------------------------------------------- #
# Define the NURBS surface class
# -------------------------------------------------------------------------------------------------------------------- #
class NurbsSurface:

    """ Create a NURBS (Non-Uniform Rational Basis Spline) surface object

        Parameters
        ----------
        control_points : ndarray with shape (ndim, n+1, m+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points (0, 1, ..., n)
            The third dimension of ´P´ spans the v-direction control points (0, 1, ..., m)

        weights : ndarray with shape (n+1, m+1)
            Array containing the weight of the control points
            The first dimension of ´W´ spans the u-direction control points weights (0, 1, ..., n)
            The second dimension of ´W´ spans the v-direction control points weights (0, 1, ..., m)

        u_degree : int
            Degree of the u-basis polynomials

        v_degree : int
            Degree of the v-basis polynomials

        u_knots : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        v_knots : ndarray with shape (s+1=m+q+2,)
            Knot vector in the v-direction
            Set the multiplicity of the first and last entries equal to ´q+1´ to obtain a clamped spline

        Notes
        -----
        This class includes methods to compute:

            - Surface coordinates for any number of dimensions
            - Analytic surface partial derivatives of any order and number of dimensions
            - Analytic mean and gaussian curvatures
            - Isoparametric curves in the u- and v-directions

        The class can be used to represent polynomial and rational Bézier, B-Spline and NURBS surfaces
        The type of surface depends on the initialization arguments

            - Polymnomial Bézier: Provide the array of control points
            - Rational Bézier:    Provide the arrays of control points and weights
            - B-Spline:           Provide the array of control points, (u,v) degrees and (u,v) knot vectors
            - NURBS:              Provide the arrays of control points and weights, (u,v) degrees and (u,v) knot vectors

        In addition, this class supports operations with real and complex numbers
        The data type used for the computations is detected from the data type of the arguments
        Using complex numbers can be useful to compute the derivative of the shape using the complex step method


        References
        ----------
        The NURBS Book. See references to equations and algorithms throughout the code
        L. Piegl and W. Tiller
        Springer, second edition

        Curves and Surfaces for CADGD. See references to equations the source code
        G. Farin
        Morgan Kaufmann Publishers, fifth edition

        All references correspond to The NURBS book unless it is explicitly stated that they come from Farin's book


    """

    def __init__(self, control_points=None, weights=None, u_degree=None, v_degree=None, u_knots=None, v_knots=None):


        # Void initialization
        if control_points is None and weights is None and u_degree is None and v_degree is None \
                and u_knots is None and v_knots is None:
            pass


        # Polynomial Bezier surface initialization
        elif weights is None and u_degree is None and u_knots is None and v_degree is None and v_knots is None:

            # Set the surface type flag
            self.surface_type = 'Bezier'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]

            # Maximum index of the control points (counting from zero)
            n = np.shape(control_points)[1] - 1
            m = np.shape(control_points)[2] - 1

            # Define the weight of the control points
            weights = np.ones((n + 1, m + 1), dtype=control_points.dtype)

            # Define the order of the basis polynomials
            u_degree = n
            v_degree = m

            # Define the knot vectors (clamped spline)
            u_knots = np.concatenate((np.zeros(u_degree), np.linspace(0, 1, n - u_degree + 2), np.ones(u_degree)))
            v_knots = np.concatenate((np.zeros(v_degree), np.linspace(0, 1, m - v_degree + 2), np.ones(v_degree)))


        # Rational Bezier surface initialization
        elif u_degree is None and u_knots is None and v_degree is None and v_knots is None:

            # Set the surface type flag
            self.surface_type = 'R-Bezier'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]

            # Maximum index of the control points (counting from zero)
            n = np.shape(control_points)[1] - 1
            m = np.shape(control_points)[2] - 1

            # Define the order of the basis polynomials
            u_degree = n
            v_degree = m

            # Define the knot vectors (clamped spline)
            u_knots = np.concatenate((np.zeros(u_degree), np.linspace(0, 1, n - u_degree + 2), np.ones(u_degree)))
            v_knots = np.concatenate((np.zeros(v_degree), np.linspace(0, 1, m - v_degree + 2), np.ones(v_degree)))


        # B-Spline surface initialization
        elif weights is None:

            # Set the surface type flag
            self.surface_type = 'B-Spline'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]

            # Maximum index of the control points (counting from zero)
            n = np.shape(control_points)[1] - 1
            m = np.shape(control_points)[2] - 1

            # Define the weight of the control points
            weights = np.ones((n + 1, m + 1), dtype=control_points.dtype)


        # NURBS surface initialization
        else:

            # Set the surface type flag
            self.surface_type = 'NURBS'

            # Set the number of dimensions of the problem
            self.ndim = np.shape(control_points)[0]


        # Declare input variables as instance variables
        self.P = control_points
        self.W = weights
        self.p = u_degree
        self.q = v_degree
        self.U = u_knots
        self.V = v_knots



    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute NURBS surface coordinates
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_value(self, u, v):

        """ Evaluate the coordinates of the surface corresponding to the (u,v) parametrization

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            u-parameter used to evaluate the surface

        v : scalar or ndarray with shape (N,)
            v-parameter used to evaluate the surface

        Returns
        -------
        S : ndarray with shape (ndim, N)
            Array containing the coordinates of the surface
            The first dimension of ´S´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´S´ spans the (u,v) parametrization sample points

        """

        # Check that u and v have the same size
        if np.isscalar(u) and np.isscalar(v): pass
        elif u.size == v.size: pass
        else: raise Exception('u and v must have the same size')

        # Evaluate the NURBS surface for the input (u,v) parametrization
        S = self.compute_nurbs_coordinates(self.P, self.W, self.p, self.q, self.U, self.V, u, v)

        return S


    @staticmethod
    def compute_nurbs_coordinates(P, W, p, q, U, V, u, v):

        """ Evaluate the coordinates of the NURBS surface corresponding to the (u,v) parametrization

        This function computes the coordinates of the NURBS surface in homogeneous space using equation 4.15 and then
        maps the coordinates to ordinary space using the perspective map given by equation 1.16. See algorithm A4.3

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1, m+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points (0, 1, ..., n)
            The third dimension of ´P´ spans the v-direction control points (0, 1, ..., m)

        W : ndarray with shape (n+1, m+1)
            Array containing the weight of the control points
            The first dimension of ´W´ spans the u-direction control points weights (0, 1, ..., n)
            The second dimension of ´W´ spans the v-direction control points weights (0, 1, ..., m)

        p : int
            Degree of the u-basis polynomials

        q : int
            Degree of the v-basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        V : ndarray with shape (s+1=m+q+2,)
            Knot vector in the v-direction
            Set the multiplicity of the first and last entries equal to ´q+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            u-parameter used to evaluate the surface

        v : scalar or ndarray with shape (N,)
            v-parameter used to evaluate the surface

        Returns
        -------
        S : ndarray with shape (ndim, N)
            Array containing the NURBS surface coordinates
            The first dimension of ´S´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´S´ spans the (u,v) parametrization sample points

        """

        # Check the shape of the input parameters
        if P.ndim > 3:           raise Exception('P must be an array of shape (ndim, n+1, m+1)')
        if W.ndim > 2:           raise Exception('W must be an array of shape (n+1, m+1)')
        if not np.isscalar(p):   raise Exception('p must be an scalar')
        if not np.isscalar(q):   raise Exception('q must be an scalar')
        if U.ndim > 1:           raise Exception('U must be an array of shape (r+1=n+p+2,)')
        if V.ndim > 1:           raise Exception('V must be an array of shape (s+1=m+q+2,)')
        if np.isscalar(u):       u = np.asarray(u)
        elif u.ndim > 1:         raise Exception('u must be a scalar or an array of shape (N,)')
        if np.isscalar(v):       v = np.asarray(v)
        elif u.ndim > 1:         raise Exception('v must be a scalar or an array of shape (N,)')

        # Shape of the array of control points
        n_dim, nn, mm = np.shape(P)

        # Highest index of the control points (counting from zero)
        n = nn - 1
        m = mm - 1

        # Compute the B-Spline basis polynomials
        N_basis_u = compute_basis_polynomials(n, p, U, u)  # shape (n+1, N)
        N_basis_v = compute_basis_polynomials(m, q, V, v)  # shape (m+1, N)

        # Map the control points to homogeneous space | P_w = (x*w,y*w,z*w,w)
        P_w = np.concatenate((P * W[np.newaxis, :], W[np.newaxis, :]), axis=0)

        # Compute the coordinates of the NURBS surface in homogeneous space
        # This implementation is vectorized to increase speed
        A = np.dot(P_w, N_basis_v)                                      # shape (ndim+1, n+1, N)
        B = np.repeat(N_basis_u[np.newaxis], repeats=n_dim+1, axis=0)   # shape (ndim+1, n+1, N)
        S_w = np.sum(A*B, axis=1)                                       # shape (ndim+1, N)

        # Map the coordinates back to the ordinary space
        S = S_w[0:-1,:]/S_w[-1, :]

        return S


    @staticmethod
    def compute_bspline_coordinates(P, p, q, U, V, u, v):

        """ Evaluate the coordinates of the B-Spline surface corresponding to the (u,v) parametrization

        This function computes the coordinates of a B-Spline surface as given by equation 3.11. See algorithm A3.5

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1, m+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points (0, 1, ..., n)
            The third dimension of ´P´ spans the v-direction control points (0, 1, ..., m)

        p : int
            Degree of the u-basis polynomials

        q : int
            Degree of the v-basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        V : ndarray with shape (s+1=m+q+2,)
            Knot vector in the v-direction
            Set the multiplicity of the first and last entries equal to ´q+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            u-parameter used to evaluate the surface

        v : scalar or ndarray with shape (N,)
            v-parameter used to evaluate the surface

        Returns
        -------
        S : ndarray with shape (ndim, N)
            Array containing the NURBS surface coordinates
            The first dimension of ´S´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´S´ spans the (u,v) parametrization sample points

        """

        # Check the shape of the input parameters
        if P.ndim > 3:           raise Exception('P must be an array of shape (ndim, n+1, m+1)')
        if not np.isscalar(p):   raise Exception('p must be an scalar')
        if not np.isscalar(q):   raise Exception('q must be an scalar')
        if U.ndim > 1:           raise Exception('U must be an array of shape (r+1=n+p+2,)')
        if V.ndim > 1:           raise Exception('V must be an array of shape (s+1=m+q+2,)')
        if np.isscalar(u):       u = np.asarray(u)
        elif u.ndim > 1:         raise Exception('u must be a scalar or an array of shape (N,)')
        if np.isscalar(v):       v = np.asarray(v)
        elif u.ndim > 1:         raise Exception('v must be a scalar or an array of shape (N,)')

        # Shape of the array of control points
        n_dim, nn, mm = np.shape(P)

        # Highest index of the control points (counting from zero)
        n = nn - 1
        m = mm - 1

        # Compute the B-Spline basis polynomials
        N_basis_u = compute_basis_polynomials(n, p, U, u)  # shape (n+1, N)
        N_basis_v = compute_basis_polynomials(m, q, V, v)  # shape (m+1, N)

        # Compute the coordinates of the B-Spline surface
        # This implementation is vectorized to increase speed
        A = np.dot(P, N_basis_v)                                        # shape (ndim, n+1, N)
        B = np.repeat(N_basis_u[np.newaxis], repeats=n_dim, axis=0)     # shape (ndim, n+1, N)
        S = np.sum(A*B,axis=1)                                          # shape (ndim, N)

        return S


    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the derivatives of the surface
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_derivative(self, u, v, order_u, order_v):

        """ Evaluate the derivative of the surface for the input u-parametrization

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            u-parameter used to evaluate the surface

        v : scalar or ndarray with shape (N,)
            v-parameter used to evaluate the surface

        order_u : int
            Order of the partial derivative in the u-direction

        order_v : int
            Order of the partial derivative in the v-direction

        Returns
        -------
        dS : ndarray with shape (ndim, N)
            Array containing the derivative of the desired order
            The first dimension of ´dC´ spans the ´(x,y,z)´ coordinates
            The second dimension of ´dC´ spans the ´u´ parametrization sample points

        """

        # Compute the array of surface derivatives up to the input (u,v) orders and slice the desired values
        dS = self.compute_nurbs_derivatives(self.P, self.W, self.p, self.q, self.U, self.V, u, v, order_u, order_v)[order_u, order_v, ...]

        return dS


    def compute_nurbs_derivatives(self, P, W, p, q, U, V, u, v, up_to_order_u, up_to_order_v):

        """ Compute the derivatives of a NURBS surface in ordinary space up to to the desired orders

        This function computes the analytic derivatives of the NURBS surface in ordinary space using equation 4.20 and
        the derivatives of the NURBS surface in homogeneous space obtained from compute_bspline_derivatives()

        The derivatives are computed recursively in a fashion similar to algorithm A4.4

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1, m+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points (0, 1, ..., n)
            The third dimension of ´P´ spans the v-direction control points (0, 1, ..., m)

        W : ndarray with shape (n+1, m+1)
            Array containing the weight of the control points
            The first dimension of ´W´ spans the u-direction control points weights (0, 1, ..., n)
            The second dimension of ´W´ spans the v-direction control points weights (0, 1, ..., m)

        p : int
            Degree of the u-basis polynomials

        q : int
            Degree of the v-basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        V : ndarray with shape (s+1=m+q+2,)
            Knot vector in the v-direction
            Set the multiplicity of the first and last entries equal to ´q+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            u-parameter used to evaluate the surface

        v : scalar or ndarray with shape (N,)
            v-parameter used to evaluate the surface

        up_to_order_u : integer
            Order of the highest derivative in the u-direction

        up_to_order_v : integer
            Order of the highest derivative in the v-direction

        Returns
        -------
        nurbs_derivatives: ndarray of shape (up_to_order_u+1, up_to_order_v+1, ndim, Nu)
            The first dimension spans the order of the u-derivatives (0, 1, 2, ...)
            The second dimension spans the order of the v-derivatives (0, 1, 2, ...)
            The third dimension spans the coordinates (x,y,z,...)
            The fourth dimension spans (u,v) parametrization sample points

        """

        # Set the data type used to initialize arrays (set `complex` if an argument is complex and `float` if not)
        u, v = np.asarray(u), np.asarray(v)
        if (P.dtype == np.complex128) or (W.dtype == np.complex128) or (u.dtype == np.complex128) or (v.dtype == np.complex128):
            data_type = np.complex128
        else:
            data_type = np.float64

        # Map the control points to homogeneous space | P_w = (x*w,y*w,z*w,w)
        P_w = np.concatenate((P * W[np.newaxis, :], W[np.newaxis, :]), axis=0)

        # Compute the derivatives of the NURBS surface in homogeneous space
        bspline_derivatives = self.compute_bspline_derivatives(P_w, p, q, U, V, u, v, up_to_order_u, up_to_order_v)
        A_ders = bspline_derivatives[:, :, 0:-1, :]
        w_ders = bspline_derivatives[:, :, [-1], :]

        # Initialize array of derivatives
        n_dim, N = np.shape(P)[0], np.size(u)
        nurbs_derivatives = np.zeros((up_to_order_u+1, up_to_order_v+1, n_dim, N), dtype=data_type)

        # Compute the derivatives of up to the desired order
        # See algorithm A4.4 from the NURBS book
        for k in range(up_to_order_u+1):
            for L in range(up_to_order_v+1):

                # Update the numerator of equation 4.20 recursively
                temp_numerator = A_ders[[k], [L], ...]

                # Summation j=0 and point_index=1:k
                for i in range(1, k + 1):
                    temp_numerator -= scipy.special.binom(k, i)*w_ders[[i], [0], ...]*nurbs_derivatives[[k-i], [L], ...]

                # Summation point_index=0 and j=1:L
                for j in range(1, L + 1):
                    temp_numerator -= scipy.special.binom(L, j)*w_ders[[0], [j], ...]*nurbs_derivatives[[k], [L-j], ...]

                # Summation point_index=1:k and j=1:L
                for i in range(1, k+1):
                    for j in range(1, L+1):
                        temp_numerator -= scipy.special.binom(k, i) * scipy.special.binom(L, j)* w_ders[[i], [j], ...] * nurbs_derivatives[[k-i], [L-j], ...]

                # Compute the (k,L)-th order NURBS surface partial derivative
                nurbs_derivatives[k, L, ...] = temp_numerator/w_ders[[0], [0], ...]

        return nurbs_derivatives


    @staticmethod
    def compute_bspline_derivatives(P, p, q, U, V, u, v, up_to_order_u, up_to_order_v):

        """ Compute the derivatives of a B-Spline (or NURBS surface in homogeneous space) up to orders
        `derivative_order_u` and `derivative_order_v`

        This function computes the analytic derivatives of a B-Spline surface using equation 3.17. See algorithm A3.6

        Parameters
        ----------
        P : ndarray with shape (ndim, n+1, m+1)
            Array containing the coordinates of the control points
            The first dimension of ´P´ spans the coordinates of the control points (any number of dimensions)
            The second dimension of ´P´ spans the u-direction control points (0, 1, ..., n)
            The third dimension of ´P´ spans the v-direction control points (0, 1, ..., m)

        p : int
            Degree of the u-basis polynomials

        q : int
            Degree of the v-basis polynomials

        U : ndarray with shape (r+1=n+p+2,)
            Knot vector in the u-direction
            Set the multiplicity of the first and last entries equal to ´p+1´ to obtain a clamped spline

        V : ndarray with shape (s+1=m+q+2,)
            Knot vector in the v-direction
            Set the multiplicity of the first and last entries equal to ´q+1´ to obtain a clamped spline

        u : scalar or ndarray with shape (N,)
            u-parameter used to evaluate the surface

        v : scalar or ndarray with shape (N,)
            v-parameter used to evaluate the surface

        up_to_order_u : integer
            Order of the highest derivative in the u-direction

        up_to_order_v : integer
            Order of the highest derivative in the v-direction

        Returns
        -------
        bspline_derivatives: ndarray of shape (up_to_order_u+1, up_to_order_v+1, ndim, Nu)
            The first dimension spans the order of the u-derivatives (0, 1, 2, ...)
            The second dimension spans the order of the v-derivatives (0, 1, 2, ...)
            The third dimension spans the coordinates (x,y,z,...)
            The fourth dimension spans (u,v) parametrization sample points

        """

        # Set the data type used to initialize arrays (set `complex` if an argument is complex and `float` if not)
        u = np.asarray(u)
        if (P.dtype == np.complex128) or (u.dtype == np.complex128) or (v.dtype == np.complex128):
            data_type = np.complex128
        else:
            data_type = np.float64

        # Set the B-Spline coordinates as the zero-th derivatives
        n_dim, N = np.shape(P)[0], np.size(u)
        bspline_derivatives = np.zeros((up_to_order_u+1, up_to_order_v+1, n_dim, N), dtype=data_type)

        # Compute the derivatives of up to the desired order
        # See algorithm A3.2 from the NURBS book
        for order_u in range(min(p, up_to_order_u) + 1):
            for order_v in range(min(q, up_to_order_v) + 1):

                # Highest index of the control points (counting from zero)
                n = np.shape(P)[1] - 1
                m = np.shape(P)[2] - 1

                # Compute the B-Spline basis polynomials
                N_basis_u = compute_basis_polynomials_derivatives(n, p, U, u, order_u)
                N_basis_v = compute_basis_polynomials_derivatives(m, q, V, v, order_v)

                # Compute the coordinates of the B-Spline surface
                # This implementation is vectorized to increase speed
                A = np.dot(P, N_basis_v)                                                # shape (ndim, n+1, N)
                B = np.repeat(N_basis_u[np.newaxis], repeats=n_dim, axis=0)             # shape (ndim, n+1, N)
                bspline_derivatives[order_u, order_v, :, :] = np.sum(A * B, axis=1)     # shape (ndim, N)

        # Note that derivatives with order higher than `p` and `q` are not computed and are be zero from initialization
        # These zero-derivatives are required to compute the higher order derivatives of rational surfaces

        return bspline_derivatives


    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the isoparametric NURBS curves
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_isocurve_u(self, u0):

        """ Create a NURBS curve object that contains the surface isoparametric curve S(u0,v)

        The isoparametric nurbs curve is defined by equations 4.16 and 4.18

        Parameters
        ----------
        u0 : scalar
            Scalar defining the u-parameter of the isoparametric curve

        Returns
        -------
        isocurve_u : instance of NurbsCurve class
            Object defining the isoparametric curve

        """

        # Map the control points to homogeneous space | P_w = (x*w,y*w,z*w,w)
        P_w = np.concatenate((self.P * self.W[np.newaxis, :], self.W[np.newaxis, :]), axis=0)

        # Compute the array of control points in homogeneous space
        n_dim, nn, mm = np.shape(P_w)
        n = nn - 1
        N_basis_u = compute_basis_polynomials(n, self.p, self.U, float(u0)).flatten()   # use float() to avoid problem with numba and integers
        N_basis_u = N_basis_u[np.newaxis, :, np.newaxis]
        Q_w = np.sum(P_w * N_basis_u, axis=1)

        # Compute the array of control points in ordinary space and the array of control point weights (inverse map)
        Q = Q_w[0:-1, :]/Q_w[-1, :]
        W = Q_w[-1, :]

        # Create the NURBS isoparametric curve in the u direction
        isocurve_u = NurbsCurve(control_points=Q, weights=W, degree=self.q, knots=self.V)

        return isocurve_u


    def get_isocurve_v(self, v0):

        """ Create a NURBS curve object that contains the surface isoparametric curve S(u,v0)

        The isoparametric nurbs curve is defined by equations 4.17 and 4.18

        Parameters
        ----------
        v0 : scalar
            Scalar defining the v-parameter of the isoparametric curve

        Returns
        -------
        isocurve_v : instance of NurbsCurve class
            Object defining the isoparametric curve

        """

        # Map the control points to homogeneous space | P_w = (x*w,y*w,z*w,w)
        P_w = np.concatenate((self.P * self.W[np.newaxis, :], self.W[np.newaxis, :]), axis=0)

        # Compute the array of control points
        n_dim, nn, mm = np.shape(P_w)
        m = mm - 1
        N_basis_v = compute_basis_polynomials(m, self.q, self.V, float(v0)).flatten()   # use float() to avoid problem with numba and integers
        N_basis_v = N_basis_v[np.newaxis, np.newaxis, :]
        Q_w = np.sum(P_w * N_basis_v, axis=2)

        # Compute the array of control points in ordinary space and the array of control point weights (inverse map)
        Q = Q_w[0:-1, :]/Q_w[-1, :]
        W = Q_w[-1, :]

        # Create the NURBS isoparametric curve in the v direction
        isocurve_v = NurbsCurve(control_points=Q, weights=W, degree=self.p, knots=self.U)

        return isocurve_v


    # ---------------------------------------------------------------------------------------------------------------- #
    # Miscellaneous methods
    # ---------------------------------------------------------------------------------------------------------------- #
    def attach_nurbs_udir(self, new_nurbs):

        """ Attatch a new NURBS curve to the end of the instance NURBS curve and return the merged NURBS curve"""

        # Check that the NURBS curves have the same degree
        if self.p != new_nurbs.p:
            raise Exception("In order to merge, the two NURBS curves must have the same degree")

        if self.q != new_nurbs.q:
            raise Exception("In order to merge, the two NURBS curves must have the same degree")

        if np.shape(self.V) != np.shape(new_nurbs.V):
            raise Exception("The two NURBS patches must have the same number of V-knots")

        if any([np.abs(v1-v2)>1e-12 for v1, v2 in zip(self.V, new_nurbs.V)]):
            raise Exception("The two NURBS patches must have the same V-knot values")

        # Combine the control points
        P = np.concatenate((self.P, new_nurbs.P), axis=1)

        # Combine the control point weights
        W = np.concatenate((self.W, new_nurbs.W), axis=0)

        # Highest index of the control points
        n1 = np.shape(self.P)[1] - 1
        n2 = np.shape(new_nurbs.P)[1] - 1

        # Combine the knot vectors (inner knot has p+1 multiplicity)
        U_start = np.zeros((self.p + 1,))
        U_end = np.ones((self.p + 1,))
        U_mid = np.ones((self.p + 1,)) / 2
        U1 = 0.00 + self.U[self.p + 1:n1 + 1] / 2
        U2 = 0.50 + new_nurbs.U[self.p + 1:n2 + 1] / 2
        U = np.concatenate((U_start, U1, U_mid, U2, U_end))

        # Create the merged NURBS surface
        mergedNurbs = NurbsSurface(control_points=P, weights=W, u_degree=self.p, v_degree=self.q, u_knots=U, v_knots=self.V)

        return mergedNurbs


    def attach_nurbs_vdir(self, new_nurbs):

        """ Attatch a new NURBS curve to the end of the instance NURBS curve and return the merged NURBS curve"""

        # Check that the NURBS curves have the same degree
        if self.p != new_nurbs.p:
            raise Exception("In order to merge, the two NURBS curves must have the same degree")

        if self.q != new_nurbs.q:
            raise Exception("In order to merge, the two NURBS curves must have the same degree")

        if np.shape(self.U) != np.shape(new_nurbs.U):
            raise Exception("The two NURBS patches must have the same number of V-knots")

        if any([np.abs(u1-u2)>1e-12 for u1, u2 in zip(self.U, new_nurbs.U)]):
            raise Exception("The two NURBS patches must have the same V-knot values")

        # Combine the control points
        P = np.concatenate((self.P, new_nurbs.P), axis=2)

        # Combine the control point weights
        W = np.concatenate((self.W, new_nurbs.W), axis=1)

        # Highest index of the control points
        n1 = np.shape(self.P)[2] - 1
        n2 = np.shape(new_nurbs.P)[2] - 1

        # Combine the knot vectors (inner knot has p+1 multiplicity)
        V_start = np.zeros((self.q + 1,))
        V_end = np.ones((self.q + 1,))
        V_mid = np.ones((self.q + 1,)) / 2
        V1 = 0.00 + self.U[self.q + 1:n1 + 1] / 2
        V2 = 0.50 + new_nurbs.U[self.q + 1:n2 + 1] / 2
        V = np.concatenate((V_start, V1, V_mid, V2, V_end))

        # Create the merged NURBS surface
        mergedNurbs = NurbsSurface(control_points=P, weights=W, u_degree=self.p, v_degree=self.q, u_knots=self.U, v_knots=V)

        return mergedNurbs


    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the unitary normal vectors
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_normals(self, u, v):

        """ Evaluate the unitary vectors normal to the surface the input (u,v) parametrization

        The definition of the unitary normal vector is given in section 19.2 (Farin's textbook)

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the normals

        v : scalar or ndarray with shape (N,)
            Scalar or array containing the v-parameter used to evaluate the normals

        Returns
        -------
        normals : ndarray with shape (ndim, N)
            Array containing the unitary vectors normal to the surface

        """

        # Compute 2 vectors tangent to the surface
        S_u = self.get_derivative(u, v, order_u=1, order_v=0)
        S_v = self.get_derivative(u, v, order_u=0, order_v=1)

        # Compute the normal vector as the cross product of the tangent vectors and normalize it
        normals = np.cross(S_u, S_v, axisa=0, axisb=0, axisc=0)
        normals = normals/np.sum(normals ** 2, axis=0) ** (1 / 2)

        return normals



    # ---------------------------------------------------------------------------------------------------------------- #
    # Compute the mean and Gaussian curvatures
    # ---------------------------------------------------------------------------------------------------------------- #
    def get_curvature(self, u, v):

        """ Evaluate the mean and gaussian curvatures of the surface the input (u,v) parametrization

        The definition of the gaussian and mean curvatures are given by equations 19.11 and 19.12 (Farin's textbook)

        Parameters
        ----------
        u : scalar or ndarray with shape (N,)
            Scalar or array containing the u-parameter used to evaluate the curvatures

        v : scalar or ndarray with shape (N,)
            Scalar or array containing the v-parameter used to evaluate the curvatures

        Returns
        -------
        mean_curvature : ndarray with shape (N, )
            Scalar or array containing the mean curvature of the surface

        gaussian_curvature : ndarray with shape (N, )
            Scalar or array containing the gaussian curvature of the surface

        """

        # Compute the partial derivatives
        S_u = self.get_derivative(u, v, order_u=1, order_v=0)
        S_v = self.get_derivative(u, v, order_u=0, order_v=1)
        S_uu = self.get_derivative(u, v, order_u=2, order_v=0)
        S_uv = self.get_derivative(u, v, order_u=1, order_v=1)
        S_vv = self.get_derivative(u, v, order_u=0, order_v=2)

        # Compute the normal vector
        N = self.get_normals(u, v)

        # Compute the components of the first fundamental form of the surface
        E = np.sum(S_u * S_u, axis=0)
        F = np.sum(S_u * S_v, axis=0)
        G = np.sum(S_v * S_v, axis=0)

        # Compute the components of the second fundamental form of the surface
        L = np.sum(S_uu * N, axis=0)
        M = np.sum(S_uv * N, axis=0)
        N = np.sum(S_vv * N, axis=0)

        # Compute the mean curvature
        mean_curvature = (1/2) * (N * E - 2 * M * F + L * G) / (E * G - F ** 2)

        # Compute the gaussian curvature
        gaussian_curvature = (L * N - M ** 2) / (E * G - F ** 2)

        return mean_curvature, gaussian_curvature



    # ---------------------------------------------------------------------------------------------------------------- #
    # Plotting functions
    # ---------------------------------------------------------------------------------------------------------------- #
    def plot(self, fig=None, ax = None,
             surface=True, surface_color='blue', colorbar=False,
             boundary=True, control_points=False, normals=False, axis_off=False, ticks_off=False,
             Nu=50, Nv=50, isocurves_u=None, isocurves_v=None):

        # Prepare the plot
        if fig is None:
            fig = mpl.pyplot.figure(figsize=(6, 5))
            ax = fig.add_subplot(111, projection='3d')
            ax.view_init(azim=-105, elev=30)
            ax.grid(False)
            ax.xaxis.pane.fill = False
            ax.yaxis.pane.fill = False
            ax.zaxis.pane.fill = False
            ax.xaxis.pane.set_edgecolor('k')
            ax.yaxis.pane.set_edgecolor('k')
            ax.zaxis.pane.set_edgecolor('k')
            ax.xaxis.pane._alpha = 0.9
            ax.yaxis.pane._alpha = 0.9
            ax.zaxis.pane._alpha = 0.9
            ax.set_xlabel('$x$ axis', fontsize=11, color='k', labelpad=18)
            ax.set_ylabel('$y$ axis', fontsize=11, color='k', labelpad=18)
            ax.set_zlabel('$z$ axis', fontsize=11, color='k', labelpad=18)
            # ax_xy.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
            # ax_xy.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
            # ax_xy.zaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
            for t in ax.xaxis.get_major_ticks(): t.label.set_fontsize(8)
            for t in ax.yaxis.get_major_ticks(): t.label.set_fontsize(8)
            for t in ax.zaxis.get_major_ticks(): t.label.set_fontsize(8)
            ax.xaxis.set_rotate_label(False)
            ax.yaxis.set_rotate_label(False)
            ax.zaxis.set_rotate_label(False)
            if ticks_off:
                ax.set_xticks([])
                ax.set_yticks([])
                ax.set_zticks([])
            if axis_off:
                ax.axis('off')

        # Add objects to the plot
        if surface:        self.plot_surface(fig, ax, color=surface_color, colorbar=colorbar, Nu=Nu, Nv=Nv)
        if boundary:       self.plot_boundary(fig, ax)
        if control_points: self.plot_control_points(fig, ax)
        if normals:        self.plot_normals(fig, ax)
        if isocurves_u:
            self.plot_isocurve_u(fig, ax, u_values=np.linspace(0, 1, isocurves_u))
        if isocurves_v:
              self.plot_isocurve_v(fig, ax, v_values=np.linspace(0, 1, isocurves_v))

        # Set the scaling of the axes
        self.rescale_plot(fig, ax)

        return fig, ax


    def plot_surface(self, fig, ax, color='blue', alpha=0.30, colorbar=False, Nu=50, Nv=50):

        # Get the surface coordinates
        u = np.linspace(0.00, 1.00, Nu)
        v = np.linspace(0.00, 1.00, Nv)
        [uu, vv] = np.meshgrid(u, v, indexing='ij')
        u = uu.flatten()
        v = vv.flatten()
        X, Y, Z = np.real(self.get_value(u, v)).reshape((3, Nu, Nv))

        # Plot the surface
        if color == 'mean_curvature':

            # Define a colormap based on the curvature values
            mean_curvature, _ = np.real(self.get_curvature(u, v))
            curvature = np.reshape(mean_curvature, (Nu, Nv))
            curvature_normalized = (curvature - np.amin(curvature)) / (np.amax(curvature) - np.amin(curvature))
            curvature_colormap = mpl.cm.viridis(curvature_normalized)

            # Plot the surface with a curvature colormap
            surf_handle = ax.plot_surface(X, Y, Z,
                                          # color='blue',
                                          # edgecolor='blue',
                                          # cmap = 'viridis',
                                          facecolors=curvature_colormap,
                                          linewidth=0.75,
                                          alpha=1,
                                          shade=False,
                                          antialiased=True,
                                          zorder=2,
                                          ccount=Nu,
                                          rcount=Nv)
            if colorbar:
                fig.set_size_inches(7, 5)
                surf_handle.set_clim(np.amin(curvature), np.amax(curvature))
                cbar = fig.colorbar(surf_handle, ax=ax, orientation='vertical', pad=0.15, fraction=0.03, aspect=20)
                cbar.set_label(color)

        elif color == 'gaussian_curvature':

            # Define a colormap based on the curvature values
            _, gaussian_curvature= np.real(self.get_curvature(u, v))
            curvature = np.reshape(gaussian_curvature, (Nu, Nv))
            curvature_normalized = (curvature - np.amin(curvature)) / (np.amax(curvature) - np.amin(curvature))
            curvature_colormap = mpl.cm.viridis(curvature_normalized)

            # Plot the surface with a curvature colormap
            surf_handle = ax.plot_surface(X, Y, Z,
                                          # color='blue',
                                          # edgecolor='blue',
                                          # cmap = 'viridis',
                                          facecolors=curvature_colormap,
                                          linewidth=0.75,
                                          alpha=1,
                                          shade=False,
                                          antialiased=True,
                                          zorder=2,
                                          ccount=Nu,
                                          rcount=Nv)
            if colorbar:
                fig.set_size_inches(7, 5)
                surf_handle.set_clim(np.amin(curvature), np.amax(curvature))
                cbar = fig.colorbar(surf_handle, ax=ax, orientation='vertical', pad=0.15, fraction=0.03, aspect=20)
                cbar.set_label(color)

        else:

            # Plot the surface with a plain color
            ax.plot_surface(X, Y, Z,
                            color=color,
                            # edgecolor='blue',
                            linewidth=0,
                            alpha=alpha,
                            shade=False,
                            antialiased=True,
                            zorder=0,
                            ccount=Nv,
                            rcount=Nu)


    def plot_boundary(self, fig, ax, color='black', linewidth=1.00, linestyle='-',
                      south=True, north=True, east=True, west=True):

        """ Plot the isoparametric curves at the boundary """

        # Create the isoparametric NURBS curves and plot them on the current figure
        if east: self.get_isocurve_u(u0=0.00).plot_curve(fig, ax, linestyle=linestyle, linewidth=linewidth, color=color)
        if west: self.get_isocurve_u(u0=1.00).plot_curve(fig, ax, linestyle=linestyle, linewidth=linewidth, color=color)
        if south: self.get_isocurve_v(v0=0.00).plot_curve(fig, ax, linestyle=linestyle, linewidth=linewidth, color=color)
        if north: self.get_isocurve_v(v0=1.00).plot_curve(fig, ax, linestyle=linestyle, linewidth=linewidth, color=color)


    def plot_isocurve_u(self, fig, ax, u_values, color='black', linewidth=1.00, linestyle='-'):

        """ Plot isoparametric curves in the u-direction """
        for u in u_values: self.get_isocurve_u(u0=u).plot_curve(fig, ax, color=color, linewidth=linewidth, linestyle=linestyle)


    def plot_isocurve_v(self, fig, ax, v_values, color='black', linewidth=1.00, linestyle='-'):

        """ Plot isoparametric curves in the v-direction """
        for v in v_values: self.get_isocurve_v(v0=v).plot_curve(fig, ax, color=color, linewidth=linewidth, linestyle=linestyle)


    def plot_control_points(self, fig, ax, color='red', linewidth=1.00, linestyle='-', markersize=5, markerstyle='o'):

        """ Plot the control points """

        # Plot the control net
        Px, Py, Pz = np.real(self.P)
        ax.plot_wireframe(Px, Py, Pz,
                          edgecolor=color,
                          linewidth=linewidth,
                          linestyles=linestyle,
                          alpha=1.0,
                          antialiased=True,
                          zorder=1)

        # Plot the control points
        points, = ax.plot(Px.flatten(), Py.flatten(), Pz.flatten())
        points.set_linewidth(linewidth)
        points.set_linestyle(' ')
        points.set_marker(markerstyle)
        points.set_markersize(markersize)
        points.set_markeredgewidth(linewidth)
        points.set_markeredgecolor(color)
        points.set_markerfacecolor('w')
        points.set_zorder(4)
        # points.set_label(' ')


    def plot_normals(self, fig, ax, number_u=10, number_v=10, scale=0.075):

        """ Plot the normal vectors """

        # Compute the surface coordinates and normal vectors
        h = 1e-6 # Add a small offset to avoid poles at the extremes [0, 1]
        u = np.linspace(0.00+h, 1.00-h, number_u)
        v = np.linspace(0.00+h, 1.00-h, number_v)
        [u, v] = np.meshgrid(u, v, indexing='xy')
        u = u.flatten()
        v = v.flatten()
        S = np.real(self.get_value(u, v))
        N = np.real(self.get_normals(u, v))

        # Scale the normal vectors and plot them
        Lu = self.get_isocurve_u(u0=0.50).get_arclength()
        Lv = self.get_isocurve_v(v0=0.50).get_arclength()
        length_scale = scale*np.real(np.amax([Lu, Lv]))
        N = length_scale * N
        ax.quiver(S[0, :], S[1, :], S[2, :], N[0, :], N[1, :], N[2, :], color='black', length=np.abs(scale), normalize=True)


    def rescale_plot(self, fig, ax):

        """ Adjust the aspect ratio of the figure """

        # Set axes aspect ratio
        ax.autoscale(enable=True)
        x_min, x_max = ax.get_xlim()
        y_min, y_max = ax.get_ylim()
        z_min, z_max = ax.get_zlim()
        x_mid = (x_min + x_max) / 2
        y_mid = (y_min + y_max) / 2
        z_mid = (z_min + z_max) / 2
        L = np.max((x_max - x_min, y_max - y_min, z_max - z_min)) / 2
        ax.set_xlim3d(x_mid - 1.0 * L, x_mid + 1.0 * L)
        ax.set_ylim3d(y_mid - 1.0 * L, y_mid + 1.0 * L)
        ax.set_zlim3d(z_mid - 1.0 * L, z_mid + 1.0 * L)

        # Adjust pad
        plt.tight_layout(pad=5.0, w_pad=None, h_pad=None)


    def plot_curvature(self, fig=None, ax=None, curvature_type='mean'):

        # Prepare the plot
        if fig is None:
            fig = mpl.pyplot.figure(figsize=(6, 5))
            ax = fig.add_subplot(111, projection='3d')
        ax.view_init(azim=-105, elev=30)
        ax.grid(False)
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False
        ax.xaxis.pane.set_edgecolor('k')
        ax.yaxis.pane.set_edgecolor('k')
        ax.zaxis.pane.set_edgecolor('k')
        ax.xaxis.pane._alpha = 0.9
        ax.yaxis.pane._alpha = 0.9
        ax.zaxis.pane._alpha = 0.9
        ax.set_xlabel('$x$ axis', fontsize=11, color='k', labelpad=18)
        ax.set_ylabel('$y$ axis', fontsize=11, color='k', labelpad=18)
        ax.set_zlabel('$z$ axis', fontsize=11, color='k', labelpad=18)
        # ax_xy.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
        # ax_xy.yaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
        # ax_xy.zaxis.set_major_formatter(mpl.ticker.FormatStrFormatter('%.1f'))
        for t in ax.xaxis.get_major_ticks(): t.label.set_fontsize(8)
        for t in ax.yaxis.get_major_ticks(): t.label.set_fontsize(8)
        for t in ax.zaxis.get_major_ticks(): t.label.set_fontsize(8)
        ax.xaxis.set_rotate_label(False)
        ax.yaxis.set_rotate_label(False)
        ax.zaxis.set_rotate_label(False)
        # ax_xy.set_xticks([])
        # ax_xy.set_yticks([])
        # ax_xy.set_zticks([])
        # ax_xy.axis('off')

        # (u,v) parametrization for the plot
        Nu, Nv = 50, 50
        u = np.linspace(0.00, 1.00, Nu)
        v = np.linspace(0.00, 1.00, Nv)
        [uu, vv] = np.meshgrid(u, v, indexing='ij')
        u = uu.flatten()
        v = vv.flatten()

        # Get the curvature
        if curvature_type == 'mean':
            curvature, _ = np.real(self.get_curvature(u, v))
        elif curvature_type == 'gaussian':
            _, curvature = np.real(self.get_curvature(u, v))
        else:
            raise Exception("Choose a valid curvature type: 'mean' or 'gaussian'")

        # Represent the curvature as a carpet plot or as a surface plot
        ax.set_xlabel('$u$', fontsize=11, color='k', labelpad=10)
        ax.set_ylabel('$v$', fontsize=11, color='k', labelpad=10)
        ax.set_zlabel(r'$\kappa$' + ' ' + curvature_type, fontsize=11, color='k', labelpad=20)
        curvature = np.reshape(curvature, (Nu, Nv))
        ax.plot_surface(uu, vv, curvature,
                        color='blue',
                        # edgecolor='blue',
                        # cmap = 'viridis',
                        # facecolors=curvature_colormap,
                        linewidth=0.75,
                        alpha=0.50,
                        shade=False,
                        antialiased=True,
                        zorder=0,
                        ccount=50,
                        rcount=50)

        # Adjust pad
        plt.tight_layout(pad=5.0, w_pad=None, h_pad=None)

        return fig, ax



# Import geometry modules
from .nurbs_basis_functions    import *
from .nurbs_curve              import *
from .nurbs_curve_circular_arc import *
from .nurbs_surface            import *
from .nurbs_surface_bilinear   import *
from .nurbs_surface_ruled      import *
from .nurbs_surface_extruded   import *
from .nurbs_surface_revolution import *
from .nurbs_surface_coons      import *



# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np

from .nurbs_surface import NurbsSurface


# -------------------------------------------------------------------------------------------------------------------- #
# Define the bilinear NURBS surface class
# -------------------------------------------------------------------------------------------------------------------- #
class NurbsSurfaceExtruded:

    """ Create the NURBS surface obtained by sweeping the NURBS curve `C(u)` a distance `d` along a direction `D`

        Create the NURBS surface given by
            S(u,v) = C(u) + v*d*D

    Parameters
    ----------
    C : NURBS curve object
        See NurbsSurface class documentation

    D : ndaray with shape (3,)
        Sweeping/extrusion direction. D may be a unitary or a non-unitary vector

    d : scalar
        Sweeping/extrusion length. d may be positive, negative or zero

    References
    ----------
    The NURBS book. Chapter 8.3
    L. Piegl and W. Tiller
    Springer, second edition

    """

    def __init__(self, C, D, d):

        # Declare input variables as instance variables
        self.C = C
        self.D = D/np.sum(D**2)**(1/2)
        self.d = d

        # Check the number of dimensions of the problem
        if np.shape(C.P)[0] != 3: raise Exception("The input NURBS must be three-dimensional")
        if np.shape(D)[0] != 3: raise Exception("The sweeping direction must be a three-dimensional vector")

        # Make the extrusion surface NURBS representation
        self.NurbsSurface = None
        self.make_Nurbs_surface()


    def make_Nurbs_surface(self):

        """ Make a NURBS surface representation of the extruded surface """

        # Define the array of control points
        P = np.concatenate((self.C.P[:, :, np.newaxis], self.C.P[:, :, np.newaxis] + self.d*self.D[:, np.newaxis, np.newaxis]), axis=2)

        # Define array of control point weights
        W = np.concatenate((self.C.W[:, np.newaxis], self.C.W[:, np.newaxis]), axis=1)

        # Maximum index of the control points (counting from zero)
        m = np.shape(P)[2] - 1

        # Define the order of the basis polynomials
        # Linear (p = 1), Quadratic (p = 2), Cubic (p = 3), etc.
        # Set p = n (number of control points minus one) to obtain a Bezier
        p = self.C.p
        q = 1

        # Define the knot vectors (clamped spline)
        # p+1 zeros, n minus p equispaced points between 0 and 1, and p+1 ones.  In total r+1 points where r=n+p+1
        # q+1 zeros, m minus q equispaced points between 0 and 1, and q+1 ones. In total s+1 points where s=m+q+1
        U = self.C.U
        V = np.concatenate((np.zeros(q), np.linspace(0, 1, m - q + 2), np.ones(q)))

        # Create the NURBS surface
        self.NurbsSurface = NurbsSurface(control_points=P, weights=W, u_degree=p, v_degree=q, u_knots=U, v_knots=V)





# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np

from .nurbs_surface import NurbsSurface


# -------------------------------------------------------------------------------------------------------------------- #
# Define the Coons NURBS surface class
# -------------------------------------------------------------------------------------------------------------------- #
class NurbsSurfaceCoons:

    """ Create the NURBS Coons surface comprised within boundaries `C_south(u)`, `C_north(u)`, `C_west(v)`, and `C_east(v)`

    Parameters
    ----------
    C_south, C_north, C_west, C_east: NURBS curve objects
        See NurbsSurface class documentation

    References
    ----------
    The NURBS book. Chapter X.Y
    L. Piegl and W. Tiller
    Springer, second edition

    """

    def __init__(self, C_south, C_north, C_west, C_east):

        # Declare input variables as instance variables
        self.C_north = C_north
        self.C_south = C_south
        self.C_west  = C_west
        self.C_east  = C_east
        self.P_north = C_north.P
        self.P_south = C_south.P
        self.P_west  = C_west.P
        self.P_east  = C_east.P
        self.W_north = C_north.W
        self.W_south = C_south.W
        self.W_west  = C_west.W
        self.W_east  = C_east.W

        # Define a tolerance to check compatibility
        tol = 1e-12

        # Check the that the NURBS curves are conforming
        # It is possible to develop a more robust implementation that can handle NURBSs with different knot vectors

        # Check the number of control points
        if any([size1 != size2 for size1, size2 in zip(np.shape(C_north.P), np.shape(C_south.P))]):
            raise Exception("The curves(C_north,C_south) must have conforming arrays of control points")

        if any([size1 != size2 for size1, size2 in zip(np.shape(C_west.P), np.shape(C_east.P))]):
            raise Exception("The curves (C_west,C_east) must have conforming arrays of control points")


        # Check the number of weights
        if any([size1 != size2 for size1, size2 in zip(np.shape(C_north.W), np.shape(C_south.W))]):
            raise Exception("The curves(C_north,C_south) must have conforming arrays of weights")

        if any([size1 != size2 for size1, size2 in zip(np.shape(C_west.W), np.shape(C_east.W))]):
            raise Exception("The curves (C_west,C_east) must have conforming arrays of weights")


        # Check the curve degrees
        if C_north.p != C_south.p:
            raise Exception("The curves(C_north,C_south) must have the same degree")

        if C_west.p != C_east.p:
            raise Exception("The curves(C_west,C_east) must have the same degree")


        # Check the knot vectors
        if np.shape(C_north.U) != np.shape(C_south.U):
            raise Exception("The curves (C_north,C_south) must have the same number of knots")

        if np.shape(C_west.U) != np.shape(C_east.U):
            raise Exception("The curves (C_west,C_east) must have the same number of knots")

        if any([np.abs(u1-u2)>tol for u1, u2 in zip(C_north.U, C_south.U)]):
            raise Exception("The curves (C_north,C_south) must have the same number of knot values")

        if any([np.abs(u1-u2)>tol for u1, u2 in zip(C_west.U, C_east.U)]):
            raise Exception("The curves (C_west,C_east) have the same number of knot values")


        # Check corner control point compatibility
        if any(np.abs(self.C_south.P[:, 0] - self.C_west.P[:, 0]) < tol):
            self.P_sw = C_south.P[:, 0]
        else:
            raise Exception("The sourth-west corner is not compatible")

        if any(np.abs(self.C_south.P[:, -1] - self.C_east.P[:, 0]) < tol):
            self.P_se = C_south.P[:, -1]
        else:
            raise Exception("The sourth-east corner is not compatible")

        if any(np.abs(self.C_north.P[:, -1] - self.C_east.P[:, -1]) < tol):
            self.P_ne = C_north.P[:, -1]
        else:
            raise Exception("The north-east corner is not compatible")

        if any(np.abs(self.C_north.P[:, 0] - self.C_west.P[:, -1]) < tol):
            self.P_nw = C_north.P[:, 0]
        else:
            raise Exception("The north-west corner is not compatible")


        # Check corner weight compatibility
        if np.abs(self.C_south.W[0] - self.C_west.W[0]) < tol:
            self.W_sw = C_south.W[0]
        else:
            raise Exception("The sourth-west weight is not compatible")

        if np.abs(self.C_south.W[-1] - self.C_east.W[0]) < tol:
            self.W_se = C_south.W[-1]
        else:
            raise Exception("The sourth-east weight is not compatible")

        if np.abs(self.C_north.W[-1] - self.C_east.W[-1]) < tol:
            self.W_ne = C_north.W[-1]
        else:
            raise Exception("The north-east weight is not compatible")

        if np.abs(self.C_north.W[0] - self.C_west.W[-1]) < tol:
            self.W_nw = C_north.W[0]
        else:
            raise Exception("The north-west weight is not compatible")


        # Make the Coons surface NURBS representation
        self.NurbsSurface = None
        self.make_nurbs_surface()



    def make_nurbs_surface(self):

        """ Make a NURBS surface representation of the Coons surface """

        # Size of the array of control points (use broadcasting to get the right shapes)
        ndim, Nu, Nv = 3, np.shape(self.P_south)[1], np.shape(self.P_west)[1]

        # Map the boundary control points to homogeneous space and set the right number of dimensions for broadcasting
        Pw_north = np.concatenate((self.P_north*self.W_north[np.newaxis, :], self.W_north[np.newaxis, :]), axis=0)[:, :, np.newaxis]
        Pw_south = np.concatenate((self.P_south*self.W_south[np.newaxis, :], self.W_south[np.newaxis, :]), axis=0)[:, :, np.newaxis]
        Pw_west  = np.concatenate((self.P_west *self.W_west [np.newaxis, :], self.W_west [np.newaxis, :]), axis=0)[:, np.newaxis, :]
        Pw_east  = np.concatenate((self.P_east *self.W_east [np.newaxis, :], self.W_east [np.newaxis, :]), axis=0)[:, np.newaxis, :]

        # Map the corner control points to homogeneous space and set the right number of dimensions for broadcasting
        Pw_sw = np.concatenate((self.P_sw*self.W_sw[np.newaxis], self.W_sw[np.newaxis]), axis=0)[:, np.newaxis, np.newaxis]
        Pw_se = np.concatenate((self.P_se*self.W_se[np.newaxis], self.W_se[np.newaxis]), axis=0)[:, np.newaxis, np.newaxis]
        Pw_ne = np.concatenate((self.P_ne*self.W_ne[np.newaxis], self.W_ne[np.newaxis]), axis=0)[:, np.newaxis, np.newaxis]
        Pw_nw = np.concatenate((self.P_nw*self.W_nw[np.newaxis], self.W_nw[np.newaxis]), axis=0)[:, np.newaxis, np.newaxis]

        # Compute the array of control points by transfinite interpolation
        u = np.linspace(0, 1, Nu)[np.newaxis, :, np.newaxis]
        v = np.linspace(0, 1, Nv)[np.newaxis, np.newaxis, :]
        term_1a = (1 - v) * Pw_south + v * Pw_north
        term_1b = (1 - u) * Pw_west + u * Pw_east
        term_2 = (1 - u) * (1 - v) * Pw_sw + u * v * Pw_ne + (1 - u) * v * Pw_nw + u * (1 - v) * Pw_se
        Pw = term_1a + term_1b - term_2

        # Compute the array of control points in ordinary space and the array of control point weights (inverse map)
        P = Pw[0:-1, :, :]/Pw[[-1], :, :]
        W = Pw[-1, :, :]

        # Define the order of the basis polynomials
        p = self.C_south.p
        q = self.C_west.p

        # Definite the knot vectors
        U = self.C_south.U
        V = self.C_west.U

        # Create the NURBS surface
        self.NurbsSurface = NurbsSurface(control_points=P, weights=W, u_degree=p, v_degree=q, u_knots=U, v_knots=V)




# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np

from .nurbs_curve import NurbsCurve


# -------------------------------------------------------------------------------------------------------------------- #
# Define the NURBS curve class
# -------------------------------------------------------------------------------------------------------------------- #
class CircularArc:

    """ Create a NURBS representation of a circular arc

        Create a NURBS representation of the circular arc parametrized by
            C(u) = O + R*cos(u)*X + R*sin(u)*Y
            theta_start <= u <= theta_end

    Parameters
    ----------

    O : ndarray with shape (ndim,)
        Center of the circular arc. ndim=2 or ndim=3

    X : ndarray with shape (ndim,)
        Unit vector lying in the plane of the circle

    Y : ndarray with shape (ndim,)
        Unit vector lying in the plane of the circle and orthogonal to X

    R : scalar
        Radius of the circle

    theta_start : scalar
        Start angle, measured with respect to X

    theta_end : scalar
        End angle, measured with respect to X

    References
    ----------
    The NURBS book. Chapter 7.5
    L. Piegl and W. Tiller
    Springer, second edition

    """

    def __init__(self, O, X, Y, R, theta_start, theta_end):

        # Set the data type used to initialize arrays (set `complex` if any argument is complex and `float` if not)
        for item in locals().values():
            data_type = np.asarray(item).dtype
            if np.issubdtype(data_type, np.complex128):
                self.data_type = np.complex128
                break
            else:
                self.data_type = np.float64

        # Declare input variables as instance variables
        self.O = O
        self.X = X/np.sum(X**2)**(1/2)
        self.Y = Y/np.sum(Y**2)**(1/2)
        self.R = R
        self.theta_start = theta_start
        self.theta_end = theta_end

        # Check the number of dimensions of the problem
        self.ndim = np.shape(O)[0]
        if not (self.ndim == 2 or self.ndim == 3):
            raise Exception("The number of dimensions of the problem must be 2 or 3")

        # Make the circular arc NURBS representation
        self.NurbsCurve = None
        self.make_nurbs_circular_arc()


    def make_nurbs_circular_arc(self):

        # Rename variables for brevity
        O = self.O
        X = self.X
        Y = self.Y
        R = self.R
        theta_start = self.theta_start
        theta_end = self.theta_end
        ndim = self.ndim

        # Correct theta_end if necessary
        if theta_end < theta_start: theta_end = theta_end + 2*np.pi

        # Angle spanned by the circular arc
        theta = theta_end - theta_start

        # Get the number of NURBS segments used to represent the circular arc
        if    theta <= 1/2*np.pi: n_arcs = 1
        elif  theta <= 2/2*np.pi: n_arcs = 2
        elif  theta <= 3/2*np.pi: n_arcs = 3
        elif  theta <= 4/2*np.pi: n_arcs = 4
        else: raise Exception('Ooops, something went wrong...')

        # Angle spanned by each segment
        delta_theta = theta/n_arcs

        # Initialize arrays of control points and weights
        n = 2*n_arcs                                        # Highest index of the control points (counting from zero)
        P = np.zeros((ndim, n + 1), dtype=self.data_type)   # Array of control points
        W = np.zeros((n + 1), dtype=self.data_type)         # Weight of the control points

        # Get the coordinates and weight of the first control point
        P0 = O + R * np.cos(theta_start) * X + R * np.sin(theta_start) * Y
        T0 = -np.sin(theta_start) * X + np.cos(theta_start) * Y
        W0 = 1.00
        P[:, 0] = P0
        W[0] = W0

        # Get the coordinates and weights of the other control points
        index, angle = 0, theta_start
        for i in range(n_arcs):

            # Angle spanned by the current segment
            angle = angle + delta_theta

            # End point of the current segment
            P2 = O + R * np.cos(angle) * X + R * np.sin(angle) * Y
            T2 = -np.sin(angle) * X + np.cos(angle) * Y

            # Solve the intersection between tangent lines T0 and T2 to compute the intermediate point
            P1 = self.intersect_lines(P0, T0, P2, T2)
            W1 = np.cos(delta_theta/2)

            # Store control points and weights
            P[:, index+1] = P1
            P[:, index+2] = P2
            W[index+1] = W1
            W[index+2] = 1

            # Get ready for the next segment!
            index = index + 2
            P0 = P2
            T0 = T2

        # Define the order of the basis polynomials
        # Linear (p = 1), Quadratic (p = 2), Cubic (p = 3), etc.
        # Set p = n (number of control points minus one) to obtain a Bezier
        p = 2

        # Knot multiplicity p+1 at the endpoints
        U = 5 + np.zeros((n + p + 2))
        U[[0, 1, 2]] = 0
        U[[-1, -2, -3]] = 1

        # Set the multiplicity of the interior knots to 2 to connect the segments
        if   n_arcs == 1: pass
        elif n_arcs == 2: U[[3, 4]] = 1/2
        elif n_arcs == 3: U[[3, 4]], U[[5, 6]]  = 1/3, 2/3
        elif n_arcs == 4: U[[3, 4]], U[[5, 6]], U[[7, 8]] = 1/4, 2/4, 3/4
        else: raise Exception('Ooops, something went wrong...')

        # Create the NURBS curve
        self.P = P
        self.W = W
        self.p = p
        self.U = U
        self.NurbsCurve = NurbsCurve(control_points=P, weights=W, degree=p, knots=U)


    def intersect_lines(self, P0, T0, P2, T2):

        # Compute the intersection after reducing the 3x2 system to a 2x2 system using the dot product
        A = np.asarray([[np.sum(T0 * T0), -np.sum(T0 * T2)], [np.sum(T0 * T2), -np.sum(T2 * T2)]])
        b = np.asarray([np.sum(P2 * T0) - np.sum(P0 * T0), np.sum(P2 * T2) - np.sum(P0 * T2)])
        u, v = np.linalg.solve(A, b)
        P1 = P0 + u * T0

        if np.sum(np.abs(((P0 + u * T0) - (P2 + v * T2)))) > 1e-12:
            raise Exception("Something went wrong computing the line intersection")

        return P1


    def plot(self):

        # Plot the NURBS curve
        fig, ax = self.NurbsCurve.plot()

        # Plot additional details (full circle, circle center and start/end points)
        if self.ndim == 2:
            u = np.linspace(0, 2 * np.pi, 100)
            P_start = np.real(self.NurbsCurve.P[:, 0])
            P_end = np.real(self.NurbsCurve.P[:, -1])
            O = np.real(self.O[:, np.newaxis])
            X = np.real(self.X[:, np.newaxis])
            Y = np.real(self.Y[:, np.newaxis])
            R = np.real(self.R)
            x, y = O + R * np.cos(u) * X + R * np.sin(u) * Y
            ax.plot(x, y, 'b')
            ax.plot([0, O[0]], [0, O[1]], 'ko-', markersize=3.5, markerfacecolor='w')
            ax.plot([O[0], P_start[0]], [O[1], P_start[1]], 'ko-', markersize=3.5, markerfacecolor='w')
            ax.plot([O[0], P_end[0]], [O[1], P_end[1]], 'ko-', markersize=3.5, markerfacecolor='w')

        elif self.ndim == 3:

            # Plot the full circle
            u = np.linspace(0, 2*np.pi, 100)
            P_start = np.real(self.NurbsCurve.P[:, 0])
            P_end = np.real(self.NurbsCurve.P[:, -1])
            O = np.real(self.O[:, np.newaxis])
            X = np.real(self.X[:, np.newaxis])
            Y = np.real(self.Y[:, np.newaxis])
            R = np.real(self.R)
            x, y, z =  O + R*np.cos(u)*X + R*np.sin(u)*Y
            ax.plot(x, y, z, 'b')
            ax.plot([0, O[0]], [0, O[1]], [0, O[2]], 'ko-', markersize=3.5, markerfacecolor='w')
            ax.plot([O[0], P_start[0]], [O[1], P_start[1]], [O[2], P_start[2]], 'ko-', markersize=3.5, markerfacecolor='w')
            ax.plot([O[0], P_end[0]], [O[1], P_end[1]], [O[2], P_end[2]], 'ko-', markersize=3.5, markerfacecolor='w')

        else:
            raise Exception('The number of dimensions must be 2 or 3')


        # Rescale the plot
        self.NurbsCurve.rescale_plot(fig, ax)

        return fig, ax










# -------------------------------------------------------------------------------------------------------------------- #
# Import packages
# -------------------------------------------------------------------------------------------------------------------- #
import numpy as np

from .nurbs_surface import NurbsSurface


# -------------------------------------------------------------------------------------------------------------------- #
# Define the bilinear NURBS surface class
# -------------------------------------------------------------------------------------------------------------------- #
class NurbsSurfaceRuled:

    """ Create the NURBS ruled surface between NURBS curves `C1(u)` and `C2(u)`

        Create the NURBS ruled surface given by
            S(u,v) = (1-v)*C1(u) + v*C2(u)

    Parameters
    ----------
    C1, C2 : NURBS curve objects
        See NurbsCurve class documentation

    References
    ----------
    The NURBS book. Chapter 8.4
    L. Piegl and W. Tiller
    Springer, second edition

    """

    def __init__(self, C1, C2):

        # Declare input variables as instance variables
        self.C1 = C1
        self.C2 = C2

        # Check the that the NURBS curves are conforming
        # It is possible to develop a more robust implementation that can handle NURBSs with different knot vectors but
        # that would require degree elevation / knot insertion (possible line of development)
        if any([size1 != size2 for size1, size2 in zip(np.shape(C1.P), np.shape(C2.P))]):
            raise Exception("The curves must have conforming arrays of control points")

        if C1.p != C2.p:
            raise Exception("The curves must have the same degree")

        if np.shape(C1.U) != np.shape(C2.U):
            raise Exception("The curves must have the same number of knots")

        if any([np.abs(u1-u2)>1e-12 for u1, u2 in zip(C1.U, C2.U)]):
            raise Exception("The curves must have the same knot values")

        # Make the ruled surface NURBS representation
        self.NurbsSurface = None
        self.make_nurbs_surface()


    def make_nurbs_surface(self):

        """ Make a NURBS surface representation of the ruled surface """

        # Define the array of control points
        P = np.concatenate((self.C1.P[:, :, np.newaxis], self.C2.P[:, :, np.newaxis]), axis=2)

        # Maximum index of the control points (counting from zero)
        m = np.shape(P)[2] - 1

        # Define array of control point weights
        W = np.concatenate((self.C1.W[:, np.newaxis], self.C2.W[:, np.newaxis]), axis=1)

        # Define the order of the basis polynomials
        # Linear (p = 1), Quadratic (p = 2), Cubic (p = 3), etc.
        # Set p = n (number of control points minus one) to obtain a Bezier
        p = self.C1.p
        q = 1

        # Define the knot vectors (clamped spline)
        # p+1 zeros, n minus p equispaced points between 0 and 1, and p+1 ones.  In total r+1 points where r=n+p+1
        # q+1 zeros, m minus q equispaced points between 0 and 1, and q+1 ones. In total s+1 points where s=m+q+1
        U = self.C1.U
        V = np.concatenate((np.zeros(q), np.linspace(0, 1, m - q + 2), np.ones(q)))

        # Create the NURBS surface
        self.NurbsSurface = NurbsSurface(control_points=P, weights=W, u_degree=p, v_degree=q, u_knots=U, v_knots=V)


